from auto_schema.diff import Diff
from auto_schema.change_type import ChangeType
import pytest

import sqlalchemy as sa
import alembic.operations.ops as alembicops

from . import conftest
from . import testingutils
from auto_schema import runner
from auto_schema import ops


class BaseTestRunner(object):

    @pytest.mark.usefixtures("empty_metadata")
    def test_compute_changes_with_empty_metadata(self, new_test_runner, empty_metadata):
        r = new_test_runner(empty_metadata)
        assert r.compute_changes() == []
        testingutils.assert_no_changes_made(r)

    @pytest.mark.usefixtures("metadata_with_table")
    def test_compute_changes_with_new_table(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        assert len(r.compute_changes()) == 1
        testingutils.assert_no_changes_made(r)

    @pytest.mark.usefixtures("metadata_with_table")
    def test_index_added_and_removed(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        r2 = testingutils.recreate_with_new_metadata(
            r, new_test_runner, metadata_with_table, conftest.metadata_with_table_with_index)

        message = r2.revision_message()
        assert message == "add index accounts_first_name_idx to accounts"

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.assert_num_tables(r2, 2)

        r3 = testingutils.recreate_metadata_fixture(
            new_test_runner, conftest.metadata_with_base_table_restored(), r2)

        message = r3.revision_message()
        assert message == "drop index accounts_first_name_idx from accounts"

        r3.run()
        testingutils.assert_num_files(r3, 3)
        testingutils.assert_num_tables(r3, 2)

    @pytest.mark.usefixtures("metadata_with_two_tables")
    def test_compute_changes_with_two_tables(self, new_test_runner, metadata_with_two_tables):
        r = new_test_runner(metadata_with_two_tables)
        assert len(r.compute_changes()) == 2
        testingutils.assert_no_changes_made(r)

    @pytest.mark.usefixtures("metadata_with_foreign_key")
    def test_compute_changes_with_foreign_key_table(self, new_test_runner, metadata_with_foreign_key):
        r = new_test_runner(metadata_with_foreign_key)
        assert len(r.compute_changes()) == 2
        testingutils.assert_no_changes_made(r)

    @pytest.mark.usefixtures("metadata_with_foreign_key_to_same_table")
    def test_compute_changes_with_foreign_key_to_same_table(self, new_test_runner, metadata_with_foreign_key_to_same_table):
        r = new_test_runner(metadata_with_foreign_key_to_same_table)
        # same table so don't need a second change. this is where the extra checks we're seeing are coming from?
        assert len(r.compute_changes()) == 1
        testingutils.assert_no_changes_made(r)

    @pytest.mark.usefixtures("metadata_with_table")
    def test_revision_message(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)

        message = r.revision_message()
        assert message == "add accounts table"

    @pytest.mark.usefixtures("metadata_with_two_tables")
    def test_revision_message_two_tables(self, new_test_runner, metadata_with_two_tables):
        r = new_test_runner(metadata_with_two_tables)

        message = r.revision_message()
        assert message == "add accounts table\nadd messages table"

    @pytest.mark.usefixtures("metadata_with_table")
    def test_new_revision(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)

        r.revision()

        # 1 schema file should have been created
        testingutils.assert_num_files(r, 1)
        testingutils.assert_num_tables(r, 0)

    @pytest.mark.usefixtures("metadata_with_table")
    def test_new_revision_with_multi_step(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)

        r.revision()

        # 1 schema file should have been created
        testingutils.assert_num_files(r, 1)

        # upgrade the schema in between. let's do a cursory check that it works
        r.upgrade()

        # confirm that 2 tables were created
        testingutils.assert_num_tables(r, 2, ['accounts', 'alembic_version'])

        r2 = testingutils.recreate_with_new_metadata(
            r, new_test_runner, metadata_with_table, conftest.messages_table)

        r2.revision()

        # we should have a 2nd schema path
        testingutils.assert_num_files(r2, 2)

        # upgrade the schema and let's confirm it works
        r2.upgrade()

        # confirm that a 3rd table was created
        testingutils.assert_num_tables(
            r2, 3, ['accounts', 'alembic_version', 'messages'])

    @pytest.mark.usefixtures("metadata_with_table")
    def test_new_table_add(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

    @pytest.mark.usefixtures("metadata_with_table")
    def test_multi_column_index_added_and_removed(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        r2 = testingutils.recreate_with_new_metadata(
            r, new_test_runner, metadata_with_table, conftest.metadata_with_multi_column_index)

        message = r2.revision_message()
        assert message == "add index accounts_first_name_last_name_idx to accounts"

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.assert_num_tables(r2, 2)

        tables = [t for t in r2.get_metadata(
        ).sorted_tables if t.name == "accounts"]
        assert len(tables) == 1
        table = tables[0]

        assert len(table.indexes) == 1
        index = table.indexes.pop()
        assert len(index.columns) == 2

        r3 = testingutils.recreate_metadata_fixture(
            new_test_runner, conftest.metadata_with_base_table_restored(), r2)

        message = r3.revision_message()
        assert message == "drop index accounts_first_name_last_name_idx from accounts"

        r3.run()
        testingutils.assert_num_files(r3, 3)
        testingutils.assert_num_tables(r3, 2)

    @pytest.mark.usefixtures("metadata_with_multi_column_pkey_constraint")
    def test_new_table_with_multi_column_pkey_constraint(self, new_test_runner, metadata_with_multi_column_pkey_constraint):
        r = new_test_runner(metadata_with_multi_column_pkey_constraint)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_with_multi_column_pkey_constraint,
            new_table_names=['user_friends_edge'],
        )

        tables = r.get_metadata().sorted_tables
        assert len(r.get_metadata().sorted_tables) == 1
        table = tables.pop()

        assert len(table.constraints) == 1
        constraint = table.constraints.pop()
        assert isinstance(constraint, sa.PrimaryKeyConstraint)
        assert len(constraint.columns) == 3

    @pytest.mark.usefixtures("metadata_with_multi_column_unique_constraint")
    def test_new_table_with_multi_column_unique_constraint(self, new_test_runner, metadata_with_multi_column_unique_constraint):
        r = new_test_runner(metadata_with_multi_column_unique_constraint)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_with_multi_column_unique_constraint,
            new_table_names=['contacts'],
        )

        tables = r.get_metadata().sorted_tables
        assert len(r.get_metadata().sorted_tables) == 1
        table = tables.pop()

        assert len(table.constraints) == 2
        constraints = table._sorted_constraints
        # first constraint, we don't care about but acknowledge
        assert isinstance(constraints[0], sa.PrimaryKeyConstraint)
        constraint = constraints[1]
        assert isinstance(constraint, sa.UniqueConstraint)
        assert len(constraint.columns) == 2

        dialect = r.get_connection().dialect.name
        # can't drop a constraint in sqlite so skipping below
        if dialect == 'sqlite':
            return

        r2 = testingutils.recreate_metadata_fixture(
            new_test_runner, conftest.metadata_with_contacts_table_with_no_unique_constraint(), r)
        message = r2.revision_message()
        assert message == "drop constraint contacts_unique_email_per_contact from contacts"

        r2.run()

        testingutils.assert_num_files(r2, 2)
        testingutils.assert_num_tables(r2, 2, ['alembic_version', 'contacts'])

    @pytest.mark.usefixtures("metadata_with_multi_column_fkey_constraint")
    def test_new_table_with_multi_column_fkey_constraint(self, new_test_runner, metadata_with_multi_column_fkey_constraint):
        r = new_test_runner(metadata_with_multi_column_fkey_constraint)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_with_multi_column_fkey_constraint,
            new_table_names=['t1', 't2'],
        )

        tables = r.get_metadata().sorted_tables
        assert len(r.get_metadata().sorted_tables) == 2
        tables = [t for t in tables if t.name == 't2']
        table = tables[0]

        assert len(table.constraints) == 2
        constraints = table._sorted_constraints
        # first constraint, we don't care about but acknowledge
        assert isinstance(constraints[0], sa.PrimaryKeyConstraint)
        constraint = constraints[1]
        assert isinstance(constraint, sa.ForeignKeyConstraint)
        assert len(constraint.columns) == 2

        dialect = r.get_connection().dialect.name
        # can't drop a constraint in sqlite so skipping below
        if dialect == 'sqlite':
            return

        r2 = testingutils.recreate_metadata_fixture(
            new_test_runner, conftest.metadata_with_multi_column_fkey_constraint_removed(), r)
        message = r2.revision_message()
        assert message == "drop constraint t2_fkey from t2"

        r2.run()

        testingutils.assert_num_files(r2, 2)
        testingutils.assert_num_tables(r2, 3, ['alembic_version', 't1', 't2'])

    # ideally we catch the expected error but this is the best we seem to be able to do do for now

    @pytest.mark.usefixtures("metadata_with_multi_column_fkey_constraint_no_constraint_reference_table")
    @pytest.mark.xfail()
    def test_new_table_with_invalid_multi_column_constraint(self, new_test_runner, metadata_with_multi_column_fkey_constraint_no_constraint_reference_table):
        r = new_test_runner(
            metadata_with_multi_column_fkey_constraint_no_constraint_reference_table)

        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_with_multi_column_fkey_constraint_no_constraint_reference_table,
            new_table_names=['t1', 't2'],
        )

    @pytest.mark.usefixtures("metadata_with_column_check_constraint")
    def test_new_table_with_column_check_constraint(self, new_test_runner, metadata_with_column_check_constraint):
        r = new_test_runner(metadata_with_column_check_constraint)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_with_column_check_constraint,
            new_table_names=['t1'],
        )

        tables = r.get_metadata().sorted_tables
        assert len(r.get_metadata().sorted_tables) == 1
        table = tables.pop()

        assert len(table.constraints) == 2
        constraints = table._sorted_constraints
        # first constraint, we don't care about but acknowledge
        assert isinstance(constraints[0], sa.PrimaryKeyConstraint)
        constraint = constraints[1]
        assert isinstance(constraint, sa.CheckConstraint)
        assert len(constraint.columns) == 0

    @pytest.mark.usefixtures("metadata_with_multi_column_check_constraint")
    def test_new_table_with_multi_column_check_constraint(self, new_test_runner, metadata_with_multi_column_check_constraint):
        r = new_test_runner(metadata_with_multi_column_check_constraint)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_with_multi_column_check_constraint,
            new_table_names=['t1'],
        )

        tables = r.get_metadata().sorted_tables
        assert len(r.get_metadata().sorted_tables) == 1
        table = tables.pop()

        assert len(table.constraints) == 4
        constraints = table._sorted_constraints
        # first constraint, we don't care about but acknowledge
        assert isinstance(constraints[0], sa.PrimaryKeyConstraint)
        for idx in range(1, 4):
            constraint = constraints[idx]
            assert isinstance(constraint, sa.CheckConstraint)
            assert len(constraint.columns) == 0

    @pytest.mark.usefixtures("metadata_with_table")
    def test_sequential_table_adds(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        # recreate runner with last runner
        r2 = testingutils.recreate_with_new_metadata(
            r, new_test_runner, metadata_with_table, conftest.messages_table)
        r2.run()

        # should have the expected files with the expected tables
        testingutils.assert_num_files(r2, 2)
        testingutils.assert_num_tables(
            r2, 3, ['accounts', 'alembic_version', 'messages'])

        testingutils.validate_metadata_after_change(r2, metadata_with_table)

    @pytest.mark.usefixtures("metadata_with_two_tables")
    def test_multiple_tables_added(self, new_test_runner, metadata_with_two_tables):
        r = new_test_runner(metadata_with_two_tables)
        r.run()

        # should have the expected file with the expected tables
        # because 2 new tables added at the same time, only one schema file needed
        testingutils.assert_num_files(r, 1)
        testingutils.assert_num_tables(
            r, 3, ['accounts', 'alembic_version', 'messages'])

        testingutils.validate_metadata_after_change(
            r, metadata_with_two_tables)

    @pytest.mark.usefixtures("metadata_with_no_edges")
    def test_no_new_edges(self, new_test_runner, metadata_with_no_edges):
        testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_no_edges,
            message='',
            num_files=1,  # just the first file
            num_changes=0
        )

    @pytest.mark.usefixtures("metadata_with_one_edge")
    def test_one_new_edge(self, new_test_runner, metadata_with_one_edge):
        testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_one_edge,
            "add edge UserToFollowersEdge"
        )

    @pytest.mark.usefixtures("metadata_with_symmetric_edge")
    def test_symmetric_edge(self, new_test_runner, metadata_with_symmetric_edge):
        testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_symmetric_edge,
            "add edge UserToFriendsEdge"
        )

    @pytest.mark.usefixtures("metadata_with_one_edge", "metadata_with_no_edges")
    def test_new_edge_then_remove(self, new_test_runner, metadata_with_one_edge, metadata_with_no_edges):
        r = testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_one_edge,
            "add edge UserToFollowersEdge"
        )

        testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_no_edges,
            "remove edge UserToFollowersEdge",
            num_files=3,
            prev_runner=r
        )

    @pytest.mark.usefixtures("metadata_with_inverse_edge")
    def test_inverse_edge_added_same_time(self, new_test_runner, metadata_with_inverse_edge, metadata_with_no_edges):
        r = testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_inverse_edge,
            "add edges UserToFolloweesEdge, UserToFollowersEdge",
        )

        testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_no_edges,
            "remove edges UserToFolloweesEdge, UserToFollowersEdge",
            prev_runner=r,
            num_files=3,
        )

    @pytest.mark.usefixtures("metadata_with_one_edge", "metadata_with_inverse_edge", "metadata_with_no_edges")
    def test_inverse_edge_added_after(self, new_test_runner, metadata_with_one_edge, metadata_with_inverse_edge, metadata_with_no_edges):
        r = testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_one_edge,
            "add edge UserToFollowersEdge"
        )

        # TODO need to test up/down for edges
        # TODO need to change the rendered to only show the minimum
        # e.g. op.remove_edge(edge_type)

        r2 = testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_inverse_edge,
            "add edge UserToFolloweesEdge\nmodify edge UserToFollowersEdge",
            prev_runner=r,
            num_files=3,
            num_changes=2,
        )

        testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_no_edges,
            "remove edges UserToFolloweesEdge, UserToFollowersEdge",
            prev_runner=r2,
            num_files=4,
            num_changes=1,
        )

    @pytest.mark.usefixtures("metadata_with_nullable_fields")
    def test_new_table_with_nullable_fields(self, new_test_runner, metadata_with_nullable_fields):
        r = new_test_runner(metadata_with_nullable_fields)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_nullable_fields)

    @pytest.mark.usefixtures("metadata_with_foreign_key_to_same_table")
    def test_with_foreign_key_to_same_table(self, new_test_runner, metadata_with_foreign_key_to_same_table):
        r = new_test_runner(metadata_with_foreign_key_to_same_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_foreign_key_to_same_table, new_table_names=["assoc_edge_config"])

    @pytest.mark.usefixtures('metadata_with_request_data')
    def test_saving_data(self, new_test_runner, metadata_with_request_data):
        r = new_test_runner(metadata_with_request_data)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_request_data, new_table_names=["request_statuses"])

        testingutils.validate_data_from_metadata(metadata_with_request_data, r)

        new_metadata = conftest.metadata_with_row_removed(
            metadata_with_request_data)
        r2 = testingutils.new_runner_from_old(
            r, new_test_runner, new_metadata)

        diff = r2.compute_changes()
        assert len(diff) == 1
        assert isinstance(diff[0], ops.RemoveRowsOp)

        assert r2.revision_message() == "remove row from request_statuses"

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.validate_metadata_after_change(r2, new_metadata)
        testingutils.validate_data_from_metadata(new_metadata, r2)

        new_metadata = conftest.metadata_with_rows_added(
            metadata_with_request_data)
        r3 = testingutils.new_runner_from_old(
            r, new_test_runner, new_metadata)

        diff = r3.compute_changes()
        assert len(diff) == 1
        assert isinstance(diff[0], ops.AddRowsOp)
        assert r3.revision_message() == "add rows to request_statuses"

        r3.run()
        testingutils.assert_num_files(r3, 3)
        testingutils.validate_metadata_after_change(r3, new_metadata)
        testingutils.validate_data_from_metadata(new_metadata, r3)

    @pytest.mark.usefixtures('metadata_with_multiple_data_tables')
    def test_saving_complex_data(self, new_test_runner, metadata_with_multiple_data_tables):
        r = new_test_runner(metadata_with_multiple_data_tables)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_with_multiple_data_tables,
            # got 3 tables
            new_table_names=["request_statuses", "rainbows"]
        )

        # data is as expected
        testingutils.validate_data_from_metadata(
            metadata_with_multiple_data_tables, r)

        # update multiple objects so there's different values
        new_metadata = conftest.metadata_with_rainbows_enum_changed(
            metadata_with_multiple_data_tables)
        r2 = testingutils.new_runner_from_old(
            r, new_test_runner, new_metadata)

        diff = r2.compute_changes()
        assert len(diff) == 1
        assert isinstance(diff[0], ops.ModifyRowsOp)
        assert r2.revision_message() == 'modify rows in rainbows'

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.validate_metadata_after_change(r2, new_metadata)
        testingutils.validate_data_from_metadata(new_metadata, r2)

    @pytest.mark.usefixtures('metadata_with_triple_pkey')
    def test_multiple_column_pkey(self, new_test_runner, metadata_with_triple_pkey):
        r = new_test_runner(metadata_with_triple_pkey)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_with_triple_pkey,
            # got 3 tables
            new_table_names=["group_members", "roles"]
        )
        testingutils.validate_data_from_metadata(metadata_with_triple_pkey, r)

        # remove rows
        new_metadata = conftest.metadata_with_triple_pkey_with_rows_removed(
            metadata_with_triple_pkey)
        r2 = testingutils.new_runner_from_old(
            r, new_test_runner, new_metadata)

        diff = r2.compute_changes()
        assert len(diff) == 1
        assert isinstance(diff[0], ops.RemoveRowsOp)
        assert r2.revision_message() == 'remove rows from group_members'

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.validate_metadata_after_change(r2, new_metadata)
        testingutils.validate_data_from_metadata(new_metadata, r2)

        # modify row
        new_metadata = conftest.metadata_with_triple_pkey_with_rows_changed(
            metadata_with_triple_pkey)
        r3 = testingutils.new_runner_from_old(
            r, new_test_runner, new_metadata)

        diff = r3.compute_changes()
        assert len(diff) == 1
        assert isinstance(diff[0], ops.ModifyRowsOp)
        assert r3.revision_message() == 'modify rows in group_members'

        r3.run()
        testingutils.assert_num_files(r3, 3)
        testingutils.validate_metadata_after_change(r3, new_metadata)
        testingutils.validate_data_from_metadata(new_metadata, r3)

    @pytest.mark.usefixtures("table_with_timestamptz_plus_date")
    def test_table_with_timestamptz_plus_date(self, new_test_runner, table_with_timestamptz_plus_date):
        r = new_test_runner(table_with_timestamptz_plus_date)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            table_with_timestamptz_plus_date,
        )
        testingutils.validate_data_from_metadata(
            table_with_timestamptz_plus_date, r)

    @pytest.mark.usefixtures("metadata_table_with_time")
    def test_table_with_time(self, new_test_runner, metadata_table_with_time):
        r = new_test_runner(metadata_table_with_time)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_table_with_time,
            new_table_names=["hours"]
        )
        testingutils.validate_data_from_metadata(metadata_table_with_time, r)

    @pytest.mark.usefixtures("metadata_table_with_timetz")
    def test_table_with_timetz(self, new_test_runner, metadata_table_with_timetz):
        r = new_test_runner(metadata_table_with_timetz)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r,
            metadata_table_with_timetz,
            new_table_names=["hours"]
        )
        testingutils.validate_data_from_metadata(metadata_table_with_timetz, r)

    @pytest.mark.usefixtures("metadata_with_one_edge")
    def test_fix_edges(self, new_test_runner, metadata_with_one_edge):
        r = testingutils.run_edge_metadata_script(
            new_test_runner,
            metadata_with_one_edge,
            "add edge UserToFollowersEdge"
        )

        # no changes when re-run
        testingutils.run_edge_metadata_script(
            new_test_runner, metadata_with_one_edge, "", num_files=2, prev_runner=r, num_changes=0)

        # can re-run with same edges and nothing happens
        runner.Runner.fix_edges(metadata_with_one_edge, {
                                'connection': r.connection})

        conn = r.get_connection()
        conn.execute('delete from assoc_edge_config')

        # validate edges fails because edges incorrect
        with pytest.raises(AssertionError):
            testingutils.validate_edges_from_metadata(
                metadata_with_one_edge, r)

        # re-run again. it fixes
        runner.Runner.fix_edges(metadata_with_one_edge, {
                                'connection': r.connection})

        # no changes
        testingutils.run_edge_metadata_script(
            new_test_runner, metadata_with_one_edge, "", num_files=2, prev_runner=r, num_changes=0)

    @pytest.mark.usefixtures("metadata_with_bigint")
    def test_tables_with_biginit(self, new_test_runner, metadata_with_bigint):
        r = new_test_runner(metadata_with_bigint)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_bigint, new_table_names=['tbl'])


class TestPostgresRunner(BaseTestRunner):

    # only in postgres because modifying columns not supported by Sqlite
    @pytest.mark.usefixtures("metadata_with_table")
    @pytest.mark.parametrize(
        "new_metadata_func, expected_message",
        [
            (conftest.metadata_with_table_text_changed,
             "modify column email_address type from VARCHAR(255) to TEXT"),
            (conftest.metadata_with_timestamp_changed,
             "modify column created_at type from DATE to TIMESTAMP"),
            (conftest.metadata_with_nullable_changed,
             "modify nullable value of column last_name from False to True"),
            (conftest.metadata_with_server_default_changed_int,
             "modify server_default value of column meaning_of_life from 42 to 35"),
            (conftest.metadata_with_server_default_changed_bool,
             "modify server_default value of column email_verified from false to TRUE"),
            (conftest.metadata_with_created_at_default_changed,
             "modify server_default value of column created_at from None to now()"),
        ])
    def test_column_attr_change(self, new_test_runner, metadata_with_table, new_metadata_func, expected_message):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        # recreate runner with last path and modified metadata
        new_metadata_func(metadata_with_table)
        r2 = new_test_runner(metadata_with_table, r)

        diff = r2.compute_changes()

        assert len(diff) == 1

        assert r2.revision_message() == expected_message

        r2.run()

        testingutils.validate_metadata_after_change(r2, metadata_with_table)

    @pytest.mark.usefixtures("address_metadata_table")
    def test_server_default_no_change_string(self, new_test_runner, address_metadata_table):
        r = new_test_runner(address_metadata_table)

        testingutils.run_and_validate_with_standard_metadata_tables(
            r, address_metadata_table, ['addresses'])

        conftest.identity_metadata_func(address_metadata_table)

        r2 = new_test_runner(address_metadata_table, r)
        diff = r2.compute_changes()

        # nothing changed, we should have no changes
        assert len(diff) == 0

    @pytest.mark.usefixtures("address_metadata_table")
    def test_server_default_change_string(self, new_test_runner, address_metadata_table):
        r = new_test_runner(address_metadata_table)

        testingutils.run_and_validate_with_standard_metadata_tables(
            r, address_metadata_table, ['addresses'])

        conftest.metadata_with_server_default_changed_string(
            address_metadata_table)

        r2 = new_test_runner(address_metadata_table, r)
        diff = r2.compute_changes()

        assert len(diff) == 1

        assert r2.revision_message() == "modify server_default value of column country from US to UK"
        r2.run()

        testingutils.validate_metadata_after_change(r2, address_metadata_table)

    @pytest.mark.usefixtures("address_metadata_table")
    def test_server_default_dropped(self, new_test_runner, address_metadata_table):
        r = new_test_runner(address_metadata_table)

        testingutils.run_and_validate_with_standard_metadata_tables(
            r, address_metadata_table, ['addresses'])

        conftest.metadata_with_server_default_dropped(
            address_metadata_table)

        r2 = new_test_runner(address_metadata_table, r)
        diff = r2.compute_changes()

        assert len(diff) == 1

        assert r2.revision_message(
        ) == "modify server_default value of column country from US to None"
        r2.run()

        testingutils.validate_metadata_after_change(r2, address_metadata_table)

    # only in postgres because "No support for ALTER of constraints in SQLite dialect"

    @pytest.mark.usefixtures("metadata_with_table")
    def test_unique_constraint_added(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        r2 = testingutils.recreate_with_new_metadata(
            r, new_test_runner, metadata_with_table, conftest.metadata_with_unique_constraint_added)

        message = r2.revision_message()
        assert message == "add unique constraint accounts_unique_email_address"

        r2.run()

        # should have the expected files with the expected tables
        testingutils.assert_num_files(r2, 2)
        testingutils.assert_num_tables(r2, 2, ['accounts', 'alembic_version'])
        testingutils.validate_metadata_after_change(r2, r2.get_metadata())

    @pytest.mark.usefixtures("metadata_with_table")
    def test_check_constraint_added_and_removed(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        r2 = testingutils.recreate_with_new_metadata(
            r, new_test_runner, metadata_with_table, conftest.metadata_with_constraint_added_after)

        message = r2.revision_message()
        assert message == "add constraint meaning_of_life_correct to accounts"

        r2.run()

        # should have the expected files with the expected tables
        testingutils.assert_num_files(r2, 2)
        testingutils.assert_num_tables(r2, 2, ['accounts', 'alembic_version'])
        testingutils.validate_metadata_after_change(r2, r2.get_metadata())

        r3 = testingutils.recreate_metadata_fixture(
            new_test_runner, conftest.metadata_with_base_table_restored(), r2)

        message = r3.revision_message()
        assert message == "drop constraint meaning_of_life_correct from accounts"

        r3.run()

        # should have the expected files with the expected tables
        testingutils.assert_num_files(r3, 3)
        testingutils.assert_num_tables(r3, 2, ['accounts', 'alembic_version'])

    @pytest.mark.usefixtures('metadata_with_enum_type')
    def test_enum_type(self, new_test_runner, metadata_with_enum_type):
        r = new_test_runner(metadata_with_enum_type)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_enum_type)

    @pytest.mark.usefixtures("metadata_with_enum_type")
    @pytest.mark.parametrize(
        'new_metadata_func, expected_diff',
        [
            (conftest.metadata_with_new_enum_value, 1),
            (conftest.metadata_with_multiple_new_enum_values, 2),
            (conftest.metadata_with_enum_value_before_first_pos, 1),
            (conftest.metadata_with_multiple_new_enum_values_at_diff_pos, 2),
            (conftest.metadata_with_multiple_new_values_before, 2),
        ]
    )
    def test_enum_additions(self, new_test_runner, metadata_with_enum_type, new_metadata_func, expected_diff):
        r = new_test_runner(metadata_with_enum_type)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_enum_type)

        # TODO this isn't ideal
        # need a good way to commit in between for separate steps in transaction to work
        conn = r.get_connection()
        conn.execute('COMMIT')

        new_metadata_func(metadata_with_enum_type)
        r2 = new_test_runner(metadata_with_enum_type, r)

        diff = r2.compute_changes()

        assert len(diff) == expected_diff

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.validate_metadata_after_change(
            r2, metadata_with_enum_type)

    @pytest.mark.usefixtures("metadata_with_enum_type")
    def test_remove_enum_value(self, new_test_runner, metadata_with_enum_type):
        r = new_test_runner(metadata_with_enum_type)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_enum_type)

        # TODO this isn't ideal
        # need a good way to commit in between for separate steps in transaction to work
        conn = r.get_connection()
        conn.execute('COMMIT')

        conftest.metadata_with_removed_enum_value(metadata_with_enum_type)
        r2 = new_test_runner(metadata_with_enum_type, r)

        with pytest.raises(ValueError, match="postgres doesn't support enum removals"):
            r2.compute_changes()

    @pytest.mark.usefixtures("metadata_with_table")
    def test_remove_column(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        new_metadata = conftest.metadata_with_removed_column()
        r2 = testingutils.new_runner_from_old(
            r, new_test_runner, new_metadata)

        diff = r2.compute_changes()

        exp_message = 'drop column meaning_of_life from table accounts'
        assert r2.revision_message() == exp_message
        assert len(diff) == 1

        d = Diff(diff)
        changes = d.changes()
        assert len(changes) == 1
        l = changes.get('accounts')
        exp_change = {
            'change': ChangeType.DROP_COLUMN,
            'desc': exp_message,
            'col': 'meaning_of_life',
        }
        assert len(l) == 1
        assert l[0] == exp_change
        d2 = Diff(diff, group_by_table=False)
        assert d2.list_changes() == [exp_change]

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.validate_metadata_after_change(r2, new_metadata)

    @pytest.mark.usefixtures("metadata_with_table")
    def test_add_multiple_cols(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        r2 = testingutils.new_runner_from_old(r, new_test_runner, conftest.metadata_with_cols_added_to_table(
            metadata_with_table))

        diff = r2.compute_changes()

        d = Diff(diff)
        changes = d.changes()
        assert len(changes) == 1
        l = changes.get('accounts')
        assert len(l) == 2
        for change in l:
            assert change['change'] == ChangeType.ADD_COLUMN
            assert change['col'] in ['new_column', 'rainbow']

    @pytest.mark.usefixtures("metadata_with_table")
    def test_new_enum_column_added_then_removed(self, new_test_runner, metadata_with_table):
        r = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        # add column with enum
        new_metadata = conftest.metadata_with_new_enum_column()
        r2 = testingutils.new_runner_from_old(
            r,
            new_test_runner,
            new_metadata
        )

        diff = r2.compute_changes()

        assert len(diff) == 2

        d = Diff(diff)
        changes = d.changes()
        assert len(changes) == 2
        l = changes.get('accounts')
        assert len(l) == 1
        assert l[0] == {
            'change': ChangeType.ADD_COLUMN,
            'desc': 'add column rainbow to table accounts',
            'col': 'rainbow',
        }

        l2 = changes.get('enum_schema')
        assert len(l2) == 1
        assert l2[0] == {
            'change': ChangeType.ADD_ENUM,
            'desc': 'add enum rainbow_type',
        }

        add_enum = [op for op in diff if isinstance(op, ops.AddEnumOp)]
        modify_table_ops = [op for op in diff if isinstance(
            op, alembicops.ModifyTableOps)]
        assert len(add_enum) == 1
        assert len(modify_table_ops) == 1

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.validate_metadata_after_change(r2, new_metadata)

        # drop column with enum
        r3 = new_test_runner(metadata_with_table, r)
        diff = r3.compute_changes()

        assert len(diff) == 2

        drop_enum = [op for op in diff if isinstance(op, ops.DropEnumOp)]
        modify_table_ops = [op for op in diff if isinstance(
            op, alembicops.ModifyTableOps)]

        assert len(drop_enum) == 1
        assert len(modify_table_ops) == 1

        r3.run()
        testingutils.assert_num_files(r3, 3)
        testingutils.validate_metadata_after_change(r3, metadata_with_table)

    @pytest.mark.usefixtures("metadata_with_enum_type")
    def test_drop_table_with_enum(self, new_test_runner, metadata_with_enum_type):
        r = new_test_runner(metadata_with_enum_type)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_enum_type)

        # no tables
        new_metadata = sa.MetaData()
        r2 = testingutils.new_runner_from_old(
            r,
            new_test_runner,
            new_metadata
        )

        diff = r2.compute_changes()

        assert len(diff) == 2

        drop_enum = [op for op in diff if isinstance(op, ops.DropEnumOp)]
        drop_table = [op for op in diff if isinstance(
            op, alembicops.DropTableOp)]
        assert len(drop_enum) == 1
        assert len(drop_table) == 1

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.validate_metadata_after_change(r2, new_metadata)

    @pytest.mark.usefixtures("metadata_with_arrays")
    def test_tables_with_arrays(self, new_test_runner, metadata_with_arrays):
        r = new_test_runner(metadata_with_arrays)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_arrays, new_table_names=['tbl'])

    @pytest.mark.usefixtures("metadata_with_json")
    def test_tables_with_json(self, new_test_runner, metadata_with_json):
        r = new_test_runner(metadata_with_json)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_json, new_table_names=['tbl'])


class TestSqliteRunner(BaseTestRunner):
    pass



from alembic.autogenerate import comparators
from . import ops
from alembic.operations import Operations, MigrateOperation
import sqlalchemy as sa
import pprint
from sqlalchemy.dialects import postgresql
import alembic.operations.ops as alembicops


@comparators.dispatch_for("schema")
def compare_edges(autogen_context, upgrade_ops, schemas):
    db_edges = {}

    for sch in schemas:

        # so first check if the table exists. if it doesn't, nothing to do here
        # TODO not using schema here either
        # https://github.com/lolopinto/ent/issues/123
        if not _table_exists(autogen_context):
            continue

        existing_edges = {}

        # get existing edges from db
        query = "SELECT * FROM assoc_edge_config"
        for row in autogen_context.connection.execute(query):
            edge = dict(row)
            existing_edges[edge['edge_name']] = edge

        db_edges[_get_schema_key(sch)] = existing_edges

    metadata_edges = autogen_context.metadata.info.setdefault("edges", {})

    # edges in metadata, but not in db, new edges that need to be added
    _process_edges(
        metadata_edges,
        db_edges,
        upgrade_ops,
        ops.AddEdgesOp,
        _meta_to_db_edge_mismatch,
    )

    # edges in db, but not in metadata, edges that need to be dropped
    _process_edges(
        db_edges,
        metadata_edges,
        upgrade_ops,
        ops.RemoveEdgesOp,
    )


def _process_edges(source_edges, compare_edges, upgrade_ops, upgrade_op, edge_mismatch_fn=None):
    alter_ops = []

    for sch, edges in source_edges.items():

        new_edges = []

        edges_for_sch = compare_edges.get(sch)
        if edges_for_sch is None:
            edges_for_sch = {}

        for k, edge in edges.items():
            compare_edge = edges_for_sch.get(edge['edge_name'])

            # for now we assume that the contents are the same. TODO, eventually support modifying edges.
            # not something we support for now
            # TODO we need modify_edge
            if compare_edge is None:
                new_edges.append(edge)
            else:
                # edge exists, let's confirm inverse_edge_type is the same
                # that's the only thing we think should change/support changing
                # convert to string to handle mismatch types e.g. str and UUID
                if str(compare_edge.get('inverse_edge_type', None)) != str(edge.get('inverse_edge_type', None)) and edge_mismatch_fn is not None:
                    alter_op = edge_mismatch_fn(edge, compare_edge, sch)
                    alter_ops.append(alter_op)
                    pass

        if len(new_edges) > 0:
            upgrade_ops.ops.append(
                upgrade_op(new_edges, schema=sch)
            )

        # do any alter operation after the add/remove edge op
        [upgrade_ops.ops.append(alter_op) for alter_op in alter_ops]


# why isn't this just metadata.sorted_tables?
def _table_exists(autogen_context):
    dialect_map = {
        'sqlite': _execute_sqlite_dialect,
        'postgresql': _execute_postgres_dialect,
    }

    dialect = autogen_context.metadata.bind.dialect.name

    if dialect_map[dialect] is None:
        raise Exception("unsupported dialect")

    return dialect_map[dialect](autogen_context.connection)


def _execute_postgres_dialect(connection):
    row = connection.execute(
        "SELECT to_regclass('%s') IS NOT NULL as exists" % (
            "assoc_edge_config")
    )
    res = row.first()
    return res['exists']


def _execute_sqlite_dialect(connection):
    row = connection.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='%s'" % (
            "assoc_edge_config")
    )
    res = row.first()
    return res is not None


def _get_schema_key(schema):
    if schema is None:
        return 'public'
    return schema


def _meta_to_db_edge_mismatch(meta_edge, db_edge, sch):
    return ops.ModifyEdgeOp(
        meta_edge['edge_type'],
        meta_edge,
        db_edge,
        schema=sch
    )


def _create_tuple_key(row, pkeys):
    l = []
    for key in pkeys:
        if not key in row:
            raise ValueError("pkey %s was not found in row" % key)
        l.append(row[key])
    return tuple(l)


@comparators.dispatch_for('schema')
def compare_data(autogen_context, upgrade_ops, schemas):
    # TODO not using schema correctly
    # https: // github.com/lolopinto/ent/issues/123

    data = autogen_context.metadata.info.setdefault("data", {})

    inspector = autogen_context.inspector
    db_metadata = sa.MetaData()
    db_metadata.reflect(inspector.bind)

    for sch in schemas:
        sch = _get_schema_key(sch)

        if not sch in data:
            continue

        schema_data = data.get(sch, {})
        for table_name in schema_data:
            table_data = schema_data[table_name]

            pkeys = table_data.get('pkeys', None)
            if pkeys is None or not isinstance(pkeys, list):
                raise ValueError("pkeys needs to be a list")

            rows = table_data.get('rows', None)
            if rows is None or not isinstance(rows, list):
                # we want list of dict...
                raise ValueError("rows needs to be a list")

            # verify that each row is valid + create tuple key
            data_rows = {_create_tuple_key(row, pkeys): row for row in rows}

            # new table. need to add
            if not table_name in db_metadata.tables:
                upgrade_ops.ops.append(ops.AddRowsOp(
                    table_name, pkeys, rows))
            else:
                _compare_db_values(autogen_context, upgrade_ops,
                                   table_name, pkeys, data_rows)


def _compare_db_values(autogen_context, upgrade_ops, table_name, pkeys, data_rows):
    connection = autogen_context.connection
    query = 'SELECT * FROM %s' % table_name

    db_rows = {}
    for row in connection.execute(query):
        d = dict(row)
        t = _create_tuple_key(d, pkeys)
        db_rows[t] = d

    deleted_rows = []
    new_rows = []
    modified_new_rows = []
    modified_old_rows = []
    for t in db_rows:
        db_row = db_rows[t]
        if t in data_rows:
            data_row = data_rows[t]
            if db_row != data_row:
                # check to see if keys are the same
                # check db since it has all the keys
                for key in db_row:
                    if db_row[key] != data_row.get(key, None):
                        modified_new_rows.append(data_row)
                        modified_old_rows.append(db_row)
                        break
        else:
            deleted_rows.append(db_row)

    for t in data_rows:
        if t not in db_rows:
            new_rows.append(data_rows[t])

    if len(new_rows) > 0:
        upgrade_ops.ops.append(ops.AddRowsOp(
            table_name, pkeys, new_rows))

    if len(deleted_rows) > 0:
        upgrade_ops.ops.append(ops.RemoveRowsOp(
            table_name, pkeys, deleted_rows))

    if len(modified_new_rows) > 0:
        if len(modified_new_rows) != len(modified_old_rows):
            raise ValueError(
                "length of modified old and new rows should be the same")

        upgrade_ops.ops.append(ops.ModifyRowsOp(
            table_name, pkeys, modified_new_rows, modified_old_rows))


@comparators.dispatch_for("schema")
def compare_schema(autogen_context, upgrade_ops, schemas):
    inspector = autogen_context.inspector

    db_metadata = sa.MetaData()
    db_metadata.reflect(inspector.bind)

    # TODO schema not being used
    # https://github.com/lolopinto/ent/issues/123
    for sch in schemas:
        conn_tables = {
            table.name: table for table in db_metadata.sorted_tables}
        metadata_tables = {
            table.name: table for table in autogen_context.metadata.sorted_tables}

        # trying to detect change in tables
        for name in conn_tables:
            if name in metadata_tables:
                _check_existing_table(
                    conn_tables[name], metadata_tables[name], upgrade_ops, sch)
            else:
                _check_removed_table(conn_tables[name], upgrade_ops, sch)

        for name in metadata_tables:
            if not name in conn_tables:
                _check_new_table(metadata_tables[name], upgrade_ops, sch)


def _check_removed_table(metadata_table, upgrade_ops, sch):
    for column in metadata_table.columns:
        _check_removed_column(column, upgrade_ops, sch)


def _check_existing_table(conn_table, metadata_table, upgrade_ops, sch):
    conn_columns = {col.name: col for col in conn_table.columns}
    metadata_columns = {
        col.name: col for col in metadata_table.columns}

    for name in conn_columns:
        if not name in metadata_columns:
            # dropped column (potentially dropped type)
            _check_removed_column(conn_columns[name], upgrade_ops, sch)
        else:
            _check_if_enum_values_changed(
                upgrade_ops, conn_columns[name], metadata_columns[name], sch)

    for name in metadata_columns:
        if not name in conn_columns:
            _check_new_column(metadata_columns[name], upgrade_ops, sch)

    conn_constraints = {
        constraint.name: constraint for constraint in conn_table.constraints}
    meta_constraints = {
        constraint.name: constraint for constraint in metadata_table.constraints}

    new_ops = []

    for name in conn_constraints:
        # no name, ignore
        # we don't do anything with None constraints and if we get here
        # it's something autogenerated or something else
        if name is None:
            continue

        constraint = conn_constraints[name]
        if not name in meta_constraints and isinstance(constraint, sa.CheckConstraint):
            if not constraint._type_bound:
                new_ops.append(
                    ops.OurDropConstraintOp.from_constraint(constraint))

    for name in meta_constraints:
        constraint = meta_constraints[name]
        if not name in conn_constraints and isinstance(constraint, sa.CheckConstraint):
            if not constraint._type_bound:
                new_ops.append(
                    ops.OurCreateCheckConstraintOp.from_constraint(constraint))

    if len(new_ops) <= 0:
        return

    existing_op = False
    for index, op in enumerate(upgrade_ops.ops):
        if not isinstance(op, alembicops.ModifyTableOps):
            continue
        if op.table_name != metadata_table.name:
            continue

        existing_op = True
        l = list(op.ops)
        # append the new ops to existing modify table ops for that table
        [l.append(new_op) for new_op in new_ops]
        # swap out the modify table ops object with new one with new ops
        upgrade_ops.ops[index] = alembicops.ModifyTableOps(
            metadata_table.name, l, schema=sch
        )
        break

    if not existing_op:
        upgrade_ops.ops.append(
            alembicops.ModifyTableOps(
                metadata_table.name, new_ops, schema=sch)
        )


def _check_new_column(metadata_column, upgrade_ops, sch):
    metadata_type = metadata_column.type
    if not isinstance(metadata_type, postgresql.ENUM):
        return

    # new column with enum type
    # time to create the type
    # adding a new type. just add to front of list
    upgrade_ops.ops.insert(
        0,
        ops.AddEnumOp(metadata_type.name, metadata_type.enums, schema=sch)
    )


def _check_removed_column(conn_column, upgrade_ops, sch):
    conn_type = conn_column.type
    if not isinstance(conn_type, postgresql.ENUM):
        return

    # column being removed. remove type also
    # we assume 1-1 for now
    upgrade_ops.ops.append(
        ops.DropEnumOp(conn_type.name, conn_type.enums, schema=sch)
    )


def _check_new_table(metadata_table, upgrade_ops, sch):
    for column in metadata_table.columns:
        _check_new_column(column, upgrade_ops, sch)


def _check_if_enum_values_changed(upgrade_ops, conn_column, metadata_column, sch):
    conn_type = conn_column.type
    metadata_type = metadata_column.type

    # not enums, bye
    if not isinstance(conn_type, postgresql.ENUM) or not isinstance(metadata_type, postgresql.ENUM):
        return

    # enums are the same, bye
    if conn_type.enums == metadata_type.enums:
        return

    conn_enums = {k: k for k in conn_type.enums}
    metadata_enums = {k: k for k in metadata_type.enums}
    for key in conn_enums:
        if key not in metadata_enums:
            raise ValueError("postgres doesn't support enum removals")

    l = len(metadata_type.enums)
    for index, value in enumerate(metadata_type.enums):
        if value not in conn_enums:
            # if not last item use BEFORE
            # options are:
            # ALTER TYPE enum_type ADD VALUE 'new_value';
            # ALTER TYPE enum_type ADD VALUE 'new_value' BEFORE 'old_value';
            # we don't need after since the previous 2 suffice so don't officially support that
            # ALTER TYPE enum_type ADD VALUE 'new_value' AFTER 'old_value';
            # only add before if previously existed
            if index != l - 1 and metadata_type.enums[index+1] in conn_enums:
                upgrade_ops.ops.append(
                    ops.AlterEnumOp(conn_type.name, value, schema=sch,
                                    before=metadata_type.enums[index + 1])
                )
            else:
                upgrade_ops.ops.append(
                    ops.AlterEnumOp(conn_type.name, value, schema=sch)
                )



import os
import sys
import argparse

# if env variable is set, manipulate the path to put local
# current directory over possibly installed auto_schema so that we
# see local changes
if os.getenv('LOCAL_AUTO_SCHEMA') == 'true':
    sys.path.insert(0, os.getcwd())


# run from auto_schema root. conflicts with pip-installed auto_schema when that exists so can't have
# that installed when runnning this...
from auto_schema.runner import Runner

from importlib import import_module

parser = argparse.ArgumentParser(
    description="generate the db schema for an ent", prog='auto_schema')
required = parser.add_argument_group('required arguments')
required.add_argument(
    '-s', '--schema', help='path to the folder the generated schema file is in', required=True)
required.add_argument(
    '-e', '--engine', help='URL to connect to the database', required=True)
parser.add_argument('-f', '--fix_edges', help='fix edges in schema into db')
parser.add_argument('-u', '--upgrade', help='upgrade')
# this is getting bad and needs to be changed soon to something that's more extensible and makes more sense
parser.add_argument('-d', '--downgrade', help='downgrade')
# only applies when downgrading, default is deleting file since it's auto created by schema
parser.add_argument('--keep_schema_files', action='store_true')
parser.add_argument('--merge_branches', action='store_true')
parser.add_argument('--history', help='alembic history', action='store_true')
parser.add_argument('--current', help='alembic current', action='store_true')
parser.add_argument('--show', help='show revision')
parser.add_argument('--heads', help='alembic heads', action='store_true')
parser.add_argument('--branches', help='alembic branches', action='store_true')
parser.add_argument('--stamp', help='alembic stamp')
parser.add_argument('--edit', help='alembic edit')
parser.add_argument('--merge', help='alembic merge')
parser.add_argument('--message', help='message if alembic merge is called')
parser.add_argument('--squash', help='squash the last N changes into one')
parser.add_argument(
    '--changes', help='get changes in schema', action='store_true')


def main():
    args = parser.parse_args()
    sys.path.append(os.path.relpath(args.schema))

    schema = import_module('schema')
    metadata = schema.get_metadata()

    if args.fix_edges:
        Runner.fix_edges(metadata, args)
    else:
        r = Runner.from_command_line(metadata, args)
        if args.upgrade is not None:
            r.upgrade(revision=args.upgrade,
                      merge_branches=args.merge_branches)
        elif args.downgrade is not None:
            r.downgrade(args.downgrade, not args.keep_schema_files)
        elif args.history is True:
            r.history()
        elif args.current is True:
            r.current()
        elif args.heads is True:
            r.heads()
        elif args.branches is True:
            r.branches()
        elif args.show is not None:
            r.show(args.show)
        elif args.stamp is not None:
            r.stamp(args.stamp)
        elif args.edit is not None:
            r.edit(args.edit)
        elif args.changes:
            r.changes()
        elif args.merge is not None:
            r.merge(args.merge, args.message)
        elif args.squash is not None:
            r.squash(args.squash)
        else:
            r.run()


if __name__ == '__main__':
    main()



import os
import pytest
import shutil
import tempfile
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from sqlalchemy.dialects import postgresql

import sqlalchemy as sa

from auto_schema import runner
from typing import List


class Postgres:
    def get_url(self, _schema_path):
        return os.getenv("DB_CONNECTION_STRING", "postgresql://localhost/autoschema_test")

    def get_finalizer(self, metadata, session, connection, transaction, engine):
        def fn():
            session.close()
            metadata.reflect()
            metadata.drop_all(bind=connection)
            # need to commit because we sometimes commit separately because of alter type ddls
            transaction.commit()

            # get any newly added enum types and drop it
            # \dT not working :(
            for row in connection.execute("select distinct pg_type.typname as enumtype from pg_type join pg_enum on pg_enum.enumtypid = pg_type.oid;"):
                connection.execute('drop type %s' % row['enumtype'])

        return fn


class SQLite:
    def get_url(self, schema_path):
        return "sqlite:///%s/%s" % (schema_path, "foo.db")
        # return "sqlite:///bar.db"  # if you want a local file to inspect for whatever reason

    def get_finalizer(self, metadata, session, connection, transaction, engine):
        def fn():
            session.close()
            transaction.rollback()
            connection.close()

        return fn


@pytest.fixture(scope="function")
def new_test_runner(request):

    def _make_new_test_runner(metadata, prev_runner=None):
        # by default, this will be none and create a temp directory where things should go
        # sometimes, when we want to test multiple revisions, we'll send the previous runner so we reuse the path
        if prev_runner is not None:
            schema_path = prev_runner.get_schema_path()
        else:
            schema_path = tempfile.mkdtemp()

        # unclear if best way but use name of class to determine postgres vs sqlite and use that
        # to make sure everything works for both
        dialect = None
        if "Postgres" in request.cls.__name__:
            dialect = Postgres()
        else:
            dialect = SQLite()

        # reuse connection if not None. same logic as schema_path above
        if prev_runner is None:
            engine = create_engine(dialect.get_url(schema_path))
            connection = engine.connect()
            metadata.bind = connection
            transaction = connection.begin()
            session = Session(bind=connection)

            request.addfinalizer(dialect.get_finalizer(
                metadata, session, connection, transaction, engine))
        else:
            connection = prev_runner.get_connection()

        r = runner.Runner(metadata, connection, schema_path)

        def delete_path():
            path = r.get_schema_path()

            # delete temp directory which was created
            if os.path.isdir(path):
                shutil.rmtree(path)

        request.addfinalizer(delete_path)

        return r

    return _make_new_test_runner


@pytest.fixture
def empty_metadata():
    metadata = sa.MetaData()
    return metadata


def default_children_of_table():
    return [
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('email_address', sa.String(255), nullable=False),
        sa.Column('first_name', sa.Text(), nullable=False),
        sa.Column('last_name', sa.Text(), nullable=False),
        sa.Column('created_at', sa.Date(), nullable=False),

        # test default sqlite
        sa.Column('meaning_of_life', sa.Integer(),
                  nullable=False, server_default='42'),
        sa.Column('email_verified', sa.Boolean(),
                  nullable=False, server_default='false'),

        sa.Column('phone_number', sa.Text(), nullable=False),
        sa.Column('updated_at', sa.TIMESTAMP(), nullable=False),
        sa.PrimaryKeyConstraint("id", name='accounts_id_pkey'),
        # support unique constraint as part of initial table creation
        sa.UniqueConstraint(
            "phone_number", name="accounts_unique_phone_number"
        ),
    ]


@pytest.fixture
def metadata_with_table():
    return metadata_with_base_table_restored()


def metadata_with_cols_added_to_table(metadata):
    return metadata_with_given_cols_added_to_table(
        metadata,
        [
            sa.Column('rainbow', sa.Text(), nullable=True),
            sa.Column('new_column', sa.Integer(), nullable=True),
        ]
    )


def metadata_with_given_cols_added_to_table(metadata: sa.MetaData, cols: List[sa.Column]):
    changes = default_children_of_table()

    [changes.append(col) for col in cols]
    metadata = sa.MetaData()
    sa.Table('accounts', metadata,
             *changes,
             )

    return metadata


def metadata_with_base_table_restored():
    metadata = sa.MetaData()
    changes = default_children_of_table()
    sa.Table('accounts', metadata,
             *changes,
             )
    return metadata


@pytest.fixture
def metadata_with_nullable_fields():
    metadata = sa.MetaData()
    sa.Table("accounts", metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('phone_number', sa.String(255), nullable=False),
             sa.Column('bio', sa.String(255), nullable=True),
             sa.Column('date_of_birth', sa.TIMESTAMP(), nullable=True),
             sa.Column('show_bio_on_profile', sa.Boolean(), nullable=True),
             sa.Column('pi', sa.Float(), nullable=True),

             sa.PrimaryKeyConstraint("id", name='accounts_id_pkey'),
             sa.UniqueConstraint(
                 "phone_number", name="accounts_unique_phone_number"),
             )
    return metadata


@pytest.fixture()
def address_metadata_table():
    metadata = sa.MetaData()
    sa.Table("addresses", metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column("street_name", sa.Text(), nullable=False),
             sa.Column("city", sa.Text(), nullable=False),
             sa.Column("zip", sa.Text(), nullable=False),
             sa.Column("apartment", sa.Text(), nullable=True),
             sa.Column("country", sa.Text(),
                       nullable=False, server_default="US"),
             sa.Column("state", sa.Text(), nullable=False),
             sa.PrimaryKeyConstraint("id", name='addresses_id_pkey'),
             )
    return metadata


@pytest.fixture()
def table_with_timestamptz_plus_date():
    metadata = sa.MetaData()
    sa.Table("accounts", metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('email_address', sa.String(255), nullable=False),
             sa.Column('first_name', sa.Text(), nullable=False),
             sa.Column('last_name', sa.Text(), nullable=False),
             sa.Column('created_at', sa.TIMESTAMP(), nullable=False),
             sa.Column('phone_number', sa.Text(), nullable=False),
             sa.Column('updated_at', sa.TIMESTAMP(
                 timezone=True), nullable=False),
             sa.PrimaryKeyConstraint("id", name='accounts_id_pkey'),
             )
    return metadata


@pytest.fixture()
def metadata_table_with_time():
    metadata = sa.MetaData()
    sa.Table("hours", metadata,
             sa.Column('day_of_week', sa.String(255), nullable=False),
             sa.Column('open', sa.Time(), nullable=False),
             sa.Column('close', sa.Time(), nullable=False),
             )
    return metadata


@pytest.fixture()
def metadata_table_with_timetz():
    metadata = sa.MetaData()
    sa.Table("hours", metadata,
             sa.Column('day_of_week', sa.String(255), nullable=False),
             sa.Column('open', sa.Time(timezone=True), nullable=False),
             sa.Column('close', sa.Time(timezone=True), nullable=False),
             )
    return metadata


@pytest.fixture
def metadata_with_arrays():
    metadata = sa.MetaData()
    sa.Table("tbl", metadata,
             sa.Column('string_list', postgresql.ARRAY(
                 sa.Text), nullable=False),
             sa.Column('string_list_2', postgresql.ARRAY(
                 sa.Text()), nullable=False),
             sa.Column('int_list', postgresql.ARRAY(
                 sa.Integer), nullable=False),
             sa.Column('bool_list', postgresql.ARRAY(
                 sa.Boolean), nullable=False),
             sa.Column('date_list', postgresql.ARRAY(sa.Date), nullable=False),
             sa.Column('time_list', postgresql.ARRAY(sa.Time), nullable=False),
             sa.Column('timetz_list', postgresql.ARRAY(
                 sa.Time(timezone=True)), nullable=False),
             sa.Column('timestamp_list', postgresql.ARRAY(
                 sa.TIMESTAMP), nullable=False),
             sa.Column('timestamptz_list', postgresql.ARRAY(
                 sa.TIMESTAMP(timezone=True)), nullable=False),
             sa.Column('float_list', postgresql.ARRAY(
                 sa.Float), nullable=False)
             )
    return metadata


@pytest.fixture
def metadata_with_json():
    metadata = sa.MetaData()
    sa.Table("tbl", metadata,
             sa.Column('jsonb', postgresql.JSONB, nullable=False),
             sa.Column('jsonb_null', postgresql.JSONB, nullable=True),
             sa.Column('json', postgresql.JSON, nullable=False),
             sa.Column('json_null', postgresql.JSON, nullable=True)
             )
    return metadata


@pytest.fixture
def metadata_with_bigint():
    metadata = sa.MetaData()
    sa.Table("tbl", metadata,
             sa.Column('bigint', sa.BigInteger, nullable=False),
             sa.Column('bigint_null', sa.BigInteger, nullable=True),
             )
    return metadata


def identity_metadata_func(metadata):
    return metadata


def metadata_with_server_default_changed_string(metadata):
    return _metadata_with_server_default_changed(metadata, 'country', 'addresses', 'UK')


def metadata_with_server_default_dropped(metadata):
    return _metadata_with_server_default_changed(metadata, 'country', 'addresses', None)


def metadata_with_unique_constraint_added(metadata):
    return _add_constraint_to_metadata(
        metadata,    # and then unique constraint added afterwards
        sa.UniqueConstraint(
            "email_address", name="accounts_unique_email_address"),
    )


# takes the account table and converts the email_address type from String(255) to Text()
def metadata_with_table_text_changed(metadata):
    return _metadata_with_col_changed(metadata, 'email_address', 'accounts', sa.Text())


# takes the account table and converts the created_at type from Date() to TIMESTAMP()
def metadata_with_timestamp_changed(metadata):
    return _metadata_with_col_changed(metadata, 'created_at', 'accounts', sa.TIMESTAMP())


# takes the account table and converts the last_name column to nullable
def metadata_with_nullable_changed(metadata):
    return _metadata_with_nullable_changed(metadata, 'last_name', 'accounts', True)


# takes the account table and converts the default value of meaning_of_life column from 42 to 35
def metadata_with_server_default_changed_int(metadata):
    return _metadata_with_server_default_changed(metadata, 'meaning_of_life', 'accounts', '35')


def metadata_with_server_default_changed_bool(metadata):
    return _metadata_with_server_default_changed(metadata, 'email_verified', 'accounts', 'TRUE')


def metadata_with_created_at_default_changed(metadata):
    return _metadata_with_server_default_changed(metadata, 'created_at', 'accounts', sa.text('now()'))


def _metadata_with_nullable_changed(metadata, col_name, table_name, nullable_value):
    def change_nullable_type(col):
        col.nullable = nullable_value
        return col

    return _apply_func_on_metadata(metadata, col_name, table_name, change_nullable_type)


def _metadata_with_server_default_changed(metadata, col_name, table_name, new_value):
    def change_server_default(col):
        # print(col, col_name, new_value, repr(new_value))
        col.server_default = new_value
        return col

    return _apply_func_on_metadata(metadata, col_name, table_name, change_server_default)


def _metadata_with_col_changed(metadata, col_name, table_name, new_type):
    def change_col_type(col):
        col.type = new_type
        return col

    return _apply_func_on_metadata(metadata, col_name, table_name, change_col_type)


def _apply_func_on_metadata(metadata, col_name, table_name, fn):
    tables = [t for t in metadata.sorted_tables if t.name == table_name]

    def apply_fn(col):
        if col.name != col_name:
            return col
        return fn(col)

    if len(tables) > 0:
        table = tables[0]
        table.columns = [apply_fn(col) for col in table.columns]

    return metadata


def metadata_with_table_with_index(metadata_with_table):
    return _add_constraint_to_metadata(
        metadata_with_table,
        # index the first name because we support searching for some reason
        sa.Index("accounts_first_name_idx", "first_name"),
    )


def metadata_with_multi_column_index(metadata_with_table):
    return _add_constraint_to_metadata(
        metadata_with_table,
        # index the first and last name because we support searching by that
        sa.Index("accounts_first_name_last_name_idx",
                 "first_name", "last_name"),
    )


@ pytest.fixture
def metadata_with_multi_column_pkey_constraint():
    metadata = sa.MetaData()
    sa.Table('user_friends_edge', metadata,
             sa.Column('id1', sa.Integer(), nullable=False),
             sa.Column('id1_type', sa.Text(), nullable=False),
             sa.Column('edge_type', sa.Integer(), nullable=False),
             sa.Column('id2', sa.Integer(), nullable=False),
             sa.Column('id2_type', sa.Text(), nullable=False),
             sa.Column('time', sa.TIMESTAMP(), nullable=False),
             sa.Column('data', sa.Text(), nullable=True),
             sa.PrimaryKeyConstraint(
                 "id1", "edge_type", "id2", name="user_friends_edge_id1_edge_type_id2_pkey"),
             )
    return metadata


@ pytest.fixture()
def metadata_with_multi_column_unique_constraint():
    metadata = metadata_with_contacts_table_with_no_unique_constraint()
    return _add_constraint_to_metadata(metadata,
                                       sa.UniqueConstraint(
                                           "email_address", "user_id", name="contacts_unique_email_per_contact"
                                       ), table_name='contacts'
                                       )


def metadata_with_contacts_table_with_no_unique_constraint():
    metadata = sa.MetaData()
    table = sa.Table('contacts', metadata,
                     sa.Column('id', sa.Integer(), nullable=False),
                     sa.Column('email_address', sa.Text(), nullable=False),
                     sa.Column('first_name', sa.Text(), nullable=False),
                     sa.Column('last_name', sa.Text(), nullable=False),
                     # ignoring the fkey here
                     sa.Column('user_id', sa.Integer(), nullable=False),
                     sa.PrimaryKeyConstraint(
                         "id", name="contacts_pkey",
                     ),
                     )
    return metadata


@ pytest.fixture()
def metadata_with_multi_column_fkey_constraint():
    metadata = sa.MetaData()
    sa.Table('t1', metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('c1', sa.Integer(), nullable=False),
             sa.Column('c2', sa.Integer(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "id", name="t1_pkey",
             ),
             sa.UniqueConstraint(
                 "c1", "c2", name="unique"
             )
             )

    sa.Table('t2', metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('c1', sa.Integer(), nullable=False),
             sa.Column('c2', sa.Integer(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "id", name="t2_pkey",
             ),
             sa.ForeignKeyConstraint(
                 ['c1', 'c2'], ['t1.c1', 't1.c2'], name="t2_fkey", ondelete="RESTRICT"
             ),
             )
    return metadata


def metadata_with_multi_column_fkey_constraint_removed():
    metadata = sa.MetaData()
    sa.Table('t1', metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('c1', sa.Integer(), nullable=False),
             sa.Column('c2', sa.Integer(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "id", name="t1_pkey",
             ),
             sa.UniqueConstraint(
                 "c1", "c2", name="unique"
             )
             )

    sa.Table('t2', metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('c1', sa.Integer(), nullable=False),
             sa.Column('c2', sa.Integer(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "id", name="t2_pkey",
             ),
             )
    return metadata


@ pytest.fixture()
def metadata_with_multi_column_fkey_constraint_no_constraint_reference_table():
    metadata = sa.MetaData()
    sa.Table('t1', metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('c1', sa.Integer(), nullable=False),
             sa.Column('c2', sa.Integer(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "id", name="t1_pkey",
             ),
             )

    sa.Table('t2', metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('c1', sa.Integer(), nullable=False),
             sa.Column('c2', sa.Integer(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "id", name="t2_pkey",
             ),
             sa.ForeignKeyConstraint(
                 ['c1', 'c2'], ['t1.c1', 't1.c2'], name="t2_fkey", ondelete="RESTRICT"
             ),
             )
    return metadata


@ pytest.fixture()
def metadata_with_column_check_constraint():
    metadata = sa.MetaData()
    sa.Table('t1', metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('price', sa.Numeric(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "id", name="t1_pkey",
             ),
             sa.CheckConstraint('price > 0', 'positive_price'),
             )
    return metadata


def metadata_with_constraint_added_after(metadata):
    return _add_constraint_to_metadata(
        metadata,
        sa.CheckConstraint('meaning_of_life = 42', 'meaning_of_life_correct'),
    )


@ pytest.fixture()
def metadata_with_multi_column_check_constraint():
    metadata = sa.MetaData()
    sa.Table('t1', metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('price', sa.Numeric(),
                       nullable=False),
             sa.Column('discounted_price', sa.Numeric(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "id", name="t1_pkey",
             ),
             # we don't do inline constraints in columns because it's harder to test since it ends up being converted from schema column constraints
             # to db table constraints which are harder to figure out so we just do everything
             # like this because easier to programmatically confirm they're the same
             sa.CheckConstraint('price > 0', 'positive_price'),
             sa.CheckConstraint('discounted_price > 0',
                                'positive_discounted_price'),
             sa.CheckConstraint('discounted_price > price', 'price_check'),
             )
    return metadata


@ pytest.fixture
def metadata_with_foreign_key_to_same_table():
    return metadata_assoc_edge_config()


@ pytest.fixture
@ pytest.mark.usefixtures("metadata_with_table")
def metadata_with_two_tables(metadata_with_table):
    messages_table(metadata_with_table)
    return metadata_with_table


@ pytest.fixture
@ pytest.mark.usefixtures("metadata_with_table")
def metadata_with_foreign_key(metadata_with_table):
    contacts_table(metadata_with_table)
    return metadata_with_table


def metadata_assoc_edge_config():
    metadata = sa.MetaData()
    assoc_edge_config_table(metadata)
    return metadata


@ pytest.fixture
def metadata_with_no_edges():
    metadata = metadata_assoc_edge_config()
    metadata.info["edges"] = {}
    return metadata


@ pytest.fixture
def metadata_with_one_edge():
    metadata = metadata_assoc_edge_config()
    # 1 edge, no inverse
    edges = {
        'public': {
            'UserToFollowersEdge': user_to_followers_edge(),
        }
    }

    metadata.info["edges"] = edges
    return metadata


@ pytest.fixture
def metadata_with_symmetric_edge():
    metadata = metadata_assoc_edge_config()
    # 1 edge, no inverse
    edges = {
        'public': {
            'UserToFriendsEdge': {
                'edge_name': 'UserToFriendsEdge',
                'edge_type': 10,
                'symmetric_edge': True,
                'edge_table': 'user_friends_edge',
            }
        }
    }

    metadata.info["edges"] = edges
    return metadata


@ pytest.fixture
def metadata_with_inverse_edge():
    metadata = metadata_assoc_edge_config()
    # 2 edges, inverse of each other
    edges = {
        'public': {
            'UserToFollowersEdge': user_to_followers_edge(inverse_edge_type=2),
            'UserToFolloweesEdge': users_to_followees_edge(inverse_edge_type=1),
        }
    }

    metadata.info["edges"] = edges
    return metadata


def default_enum_values():
    return ['open', 'pending', 'closed']


def status_table_info(l):
    return {
        'pkeys': ['status'],
        'rows': [{'status': v} for v in l]
    }


@ pytest.fixture
def metadata_with_request_data():
    metadata = metdata_enum_table()
    data = {
        'public': {
            'request_statuses': status_table_info(default_enum_values())
        }
    }
    metadata.info["data"] = data
    return metadata


def metadata_with_row_removed(metadata):
    enums = default_enum_values()
    enums.remove('open')

    data = {
        'public': {
            'request_statuses': status_table_info(enums)
        }
    }
    metadata.info["data"] = data
    return metadata


def metadata_with_rows_added(metadata):
    enums = default_enum_values()
    enums.remove('open')  # putting this here since done from above
    enums.append('initial')  # e.g. open -> initial
    enums.append('reviewed')

    data = {
        'public': {
            'request_statuses': status_table_info(enums)
        }
    }
    metadata.info["data"] = data
    return metadata


@ pytest.fixture
def metadata_with_multiple_data_tables():
    metadata = metdata_enum_table()
    complex_enum_table(metadata)

    data = {
        'public': {
            'request_statuses': status_table_info(default_enum_values()),
            'rainbows': {
                'pkeys': ['color'],
                'rows': [
                    {'color': 'red', 'description': "Red"},
                    {'color': 'orange', 'description': "Orange"},
                    {'color': 'yellow', 'description': "Yellow"},
                    {'color': 'green', 'description': "Green"},
                    {'color': 'blue', 'description': "Blue"},
                    {'color': 'indigo', 'description': "Indigo"},
                    {'color': 'violet', 'description': "Violet"},
                ]
            }
        }
    }
    metadata.info["data"] = data
    return metadata


def metadata_with_rainbows_enum_changed(metadata):
    data = metadata.info['data']
    rows = data['public']['rainbows']['rows']
    rows[3]['description'] = 'Really Green'
    rows[4]['description'] = 'Navy Blue'

    return metadata


def metdata_enum_table():
    metadata = sa.MetaData()
    enum_table(metadata)
    return metadata


def enum_table(metadata):
    sa.Table('request_statuses', metadata,
             sa.Column('status', sa.String(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "status", name="request_status_pkey"),
             )


def complex_enum_table(metadata):
    sa.Table('rainbows', metadata,
             sa.Column('color', sa.String(), nullable=False),
             sa.Column('description', sa.String(), nullable=False),
             sa.PrimaryKeyConstraint('color', name='rainbows_color_pkey')
             )


def roles_table(metadata):
    sa.Table('roles', metadata,
             sa.Column('role', sa.String(), nullable=False),
             sa.PrimaryKeyConstraint('role', name='roles_role_pkey'))


def group_members_table(metadata):
    sa.Table('group_members', metadata,
             sa.Column('group_id', sa.Integer(), nullable=False),
             sa.Column('user_id', sa.Integer(), nullable=False),
             sa.Column('role', sa.String(), nullable=False),
             sa.Column('data', sa.Text(), nullable=True),
             # triple primary key!
             # with a foreign key also
             sa.PrimaryKeyConstraint(
                 'group_id', 'user_id', 'role', name='group_member_roles_pkey'),
             sa.ForeignKeyConstraint(['role'], [
                                     'roles.role'], name="roles_role_fkey", ondelete="CASCADE"),
             )


def roles_table_info():
    return {
        'pkeys': ['role'],
        'rows': [{'role': v} for v in ['admin', 'member', 'archived_member']]
    }


@ pytest.fixture()
def metadata_with_triple_pkey():
    metadata = sa.MetaData()
    roles_table(metadata)
    group_members_table(metadata)

    data = {
        'public': {
            'roles': roles_table_info(),
            'group_members': {
                'pkeys': ['group_id', 'user_id', 'role'],
                'rows': [
                    {
                        'group_id': 1, 'user_id': 100, 'role': 'admin',
                    },
                    {
                        'group_id': 1, 'user_id': 200, 'role': 'member',
                    },
                    {
                        'group_id': 1, 'user_id': 200, 'role': 'admin',
                    },
                ]
            }
        }
    }
    metadata.info['data'] = data
    return metadata


def metadata_with_triple_pkey_with_rows_removed(metadata):
    data = {
        'public': {
            'roles': roles_table_info(),
            'group_members': {
                'pkeys': ['group_id', 'user_id', 'role'],
                'rows': [
                    {
                        'group_id': 1, 'user_id': 100, 'role': 'admin',
                    },
                    # remove everything with user 200
                ]
            }
        }
    }
    metadata.info['data'] = data

    return metadata


def metadata_with_triple_pkey_with_rows_changed(metadata):
    data = {
        'public': {
            'roles': roles_table_info(),
            'group_members': {
                'pkeys': ['group_id', 'user_id', 'role'],
                'rows': [
                    {
                        # add data which wasn't previously there
                        'group_id': 1, 'user_id': 100, 'role': 'admin', 'data': '123'
                    },
                ]
            }
        }
    }
    metadata.info['data'] = data

    return metadata


@ pytest.fixture
def metadata_with_enum_type():
    metadata = sa.MetaData()

    rainbow = ('red', 'orange', 'yellow', 'green', 'blue', 'indigo', 'violet')
    enum = postgresql.ENUM(*rainbow, name='rainbow_type')

    sa.Table('accounts', metadata,
             sa.Column('id', sa.Integer(), nullable=False),
             sa.Column('rainbow', enum, nullable=False),
             sa.PrimaryKeyConstraint("id", name='accounts_id_pkey'),
             )
    return metadata


def _apply_func_on_enum(metadata, fn):
    return _apply_func_on_metadata(metadata, 'rainbow', 'accounts', fn)


def metadata_with_new_enum_value(metadata_with_enum):
    def add_purple(col):
        col.type.enums.append('purple')
        return col

    return _apply_func_on_enum(metadata_with_enum, add_purple)


def metadata_with_enum_value_before_first_pos(metadata_with_enum):
    def insert_purple(col):
        col.type.enums.insert(0, 'purple')
        return col

    return _apply_func_on_enum(metadata_with_enum, insert_purple)


def metadata_with_multiple_new_values_before(metadata_with_enum):
    def insert_colors(col):
        col.type.enums.insert(2, 'purple')
        col.type.enums.insert(4, 'black')
        return col

    return _apply_func_on_enum(metadata_with_enum, insert_colors)


def metadata_with_multiple_new_enum_values(metadata_with_enum):
    def append_colors(col):
        col.type.enums.append('purple')
        col.type.enums.append('black')
        return col

    return _apply_func_on_enum(metadata_with_enum, append_colors)


def metadata_with_multiple_new_enum_values_at_diff_pos(metadata_with_enum):
    def change_colors(col):
        col.type.enums.insert(3, 'purple')
        col.type.enums.append('black')
        return col

    return _apply_func_on_enum(metadata_with_enum, change_colors)


def metadata_with_removed_enum_value(metadata_with_enum):
    def remove_color(col):
        col.type.enums.remove('green')
        return col

    return _apply_func_on_enum(metadata_with_enum, remove_color)


def metadata_with_removed_column():
    changes = default_children_of_table()

    cols = [c for c in changes if isinstance(
        c, sa.Column) and c.name == 'meaning_of_life']
    col = cols[0]
    changes.remove(col)

    metadata = sa.MetaData()
    sa.Table('accounts', metadata,
             *changes,
             )
    return metadata


def metadata_with_new_enum_column():
    changes = default_children_of_table()

    rainbow = ('red', 'orange', 'yellow', 'green', 'blue', 'indigo', 'violet')
    enum = postgresql.ENUM(*rainbow, name='rainbow_type')

    changes.append(
        sa.Column('rainbow', enum, nullable=False)
    )
    metadata = sa.MetaData()
    sa.Table('accounts', metadata,
             *changes,
             )

    return metadata


def user_to_followers_edge(edge_type=1, inverse_edge_type=None):
    return {
        'edge_name': 'UserToFollowersEdge',
        'edge_type': edge_type,
        'edge_table': 'user_followers_edge',
        'symmetric_edge': False,
        'inverse_edge_type': inverse_edge_type,
    }


def users_to_followees_edge(edge_type=2, inverse_edge_type=None):
    return {
        'edge_name': 'UserToFolloweesEdge',
        'edge_type': edge_type,
        'edge_table': 'user_followers_edge',
        'symmetric_edge': False,
        'inverse_edge_type': inverse_edge_type,
    }


def messages_table(metadata):
    sa.Table('messages', metadata,
             sa.Column('id', sa.Integer()),
             sa.Column('thread_id', sa.Integer(), nullable=False),
             sa.Column('message', sa.Text(), nullable=False),
             sa.PrimaryKeyConstraint("id", name='messages_id_pkey'),
             )


def contacts_table(metadata):
    sa.Table('contacts', metadata,
             sa.Column('id', sa.Integer(), primary_key=True),
             sa.Column('account_id', sa.Integer, nullable=False),
             sa.Column('name', sa.String(255), nullable=False),
             sa.PrimaryKeyConstraint("id", name="contacts_id_pkey"),
             sa.ForeignKeyConstraint(['account_id'], [
                                     'accounts.id'], name="contacts_account_id_fkey", ondelete="CASCADE"),
             )


# this is the structure that's automatically created by ent
# TODO: should test this so that if the generation changes, we catch this
# need an integration test of some sort
# this doesn't quite match because i'm doing integer instead of uuid
# todo: one postgres test with uuid
def assoc_edge_config_table(metadata):
    sa.Table('assoc_edge_config', metadata,
             sa.Column('edge_type', sa.Integer(), nullable=False,
                       sqlite_on_conflict_primary_key='IGNORE'),
             sa.Column('edge_name', sa.Text(), nullable=False),
             # use false instead of FALSE to avoid the need for craziness here
             sa.Column('symmetric_edge', sa.Boolean(),
                       nullable=False, server_default='false'),
             sa.Column('inverse_edge_type', sa.Integer(), nullable=True),
             sa.Column('edge_table', sa.Text(), nullable=False),
             sa.Column('created_at', sa.TIMESTAMP(), nullable=False),
             sa.Column('updated_at', sa.TIMESTAMP(), nullable=False),
             sa.PrimaryKeyConstraint(
                 "edge_type", name="assoc_edge_config_edge_type_pkey"),
             sa.ForeignKeyConstraint(['inverse_edge_type'], ['assoc_edge_config.edge_type'],
                                     name="assoc_edge_config_inverse_edge_type_fkey", ondelete="RESTRICT"),
             )


def _add_constraint_to_metadata(metadata, constraint, table_name="accounts"):
    tables = [t for t in metadata.sorted_tables if t.name == table_name]
    table = tables[0]

    table.append_constraint(constraint)
    return metadata



import os

from alembic.config import Config
from alembic import command
from alembic.script import ScriptDirectory

from . import runner
from typing import TypedDict


class UpgradeInfo(TypedDict):
    unmerged_branches: bool
    merged_and_upgraded_head: bool


class Command(object):

    def __init__(self, connection, schema_path):
        alembic_cfg = Config()

        # script location is where we're running this from so keep that local
        alembic_cfg.set_main_option(
            "script_location", os.path.dirname(__file__))
        # print("env.py location", os.path.dirname(__file__))

        alembic_cfg.set_main_option(
            "version_locations", os.path.join(schema_path, "versions"))

        # should probably make some of these configurable eventually
        alembic_cfg.set_main_option(
            "file_template", "%%(rev)s_%%(year)d%%(month)d%%(day)d%%(hour)d%%(minute)d%%(second)d_%%(slug)s")
        alembic_cfg.set_main_option("truncate_slug_length", "40")
        alembic_cfg.set_main_option("timezone", "utc")  # use utc timezone.
        alembic_cfg.set_section_option(
            "alembic:exclude", "tables", runner.Runner.exclude_tables())

        # for default formatting
        alembic_cfg.set_section_option('post_write_hooks', 'hooks', 'autopep8')
        alembic_cfg.set_section_option(
            'post_write_hooks', 'autopep8.type', 'console_scripts')
        alembic_cfg.set_section_option(
            'post_write_hooks', 'autopep8.entrypoint', 'autopep8')
        alembic_cfg.set_section_option(
            'post_write_hooks', 'autopep8.options', '--in-place')

        self.alembic_cfg = alembic_cfg

        # pass connection instead of re-creating it and using a sqlalchemy_url file
        alembic_cfg.attributes['connection'] = connection

    # Returns the current revision of the database. Same as calling `alembic current`
    def current(self):
        command.current(self.alembic_cfg, verbose=True)

    # Simulates running the `alembic revision -m` command
    def revision(self, message):
        heads = self.get_heads()
        if len(heads) > 1:
            # if multiple heads, now safe to do a merge first before the revision since we can't
            # keep making (automatic) changes in a branch
            merge_message = self._get_merge_message(heads)
            # create a merge revision and upgrade to it
            command.merge(self.alembic_cfg, 'heads', message=merge_message)
            command.upgrade(self.alembic_cfg, 'head')

        command.revision(self.alembic_cfg, message, autogenerate=True)

    def get_heads(self):
        script = ScriptDirectory.from_config(self.alembic_cfg)

        return script.get_heads()

    def _get_merge_message(self, heads) -> str:
        return 'merge revisions %s ' % ", ".join(heads)

    # Simulates running the `alembic upgrade` command
    def upgrade(self, revision='head', merge_branches=False) -> UpgradeInfo:
        ret: UpgradeInfo = {}
        merge_message = None
        if revision == 'head':
            # check for current heads
            # if more than one, update to heads
            # and then also create merge script
            heads = self.get_heads()

            if len(heads) > 1:
                if merge_branches:
                    merge_message = self._get_merge_message(heads)
                else:
                    ret['unmerged_branches'] = True

                # need to change to upgrade to heads and then merge and upgrade that to head
                revision = 'heads'

        command.upgrade(self.alembic_cfg, revision)

        if merge_message is None:
            return ret

        # merge the heads
        command.merge(self.alembic_cfg, 'heads', message=merge_message)
        # upgrade to head
        command.upgrade(self.alembic_cfg, 'head')
        # we want to flag this back somehow so that developer knows what happened
        ret['merged_and_upgraded_head'] = True
        return ret

    # Simulates running the `alembic downgrade` command
    def downgrade(self, revision='', delete_files=True):
        paths = []
        if delete_files:
            paths = self._get_paths_to_delete(revision)
        command.downgrade(self.alembic_cfg, revision)

        # if downgrade worked, delete files
        location = self.alembic_cfg.get_main_option('version_locations')
        for path in paths:
            os.remove(os.path.join(location, path))

    def _get_paths_to_delete(self, revision):
        script = ScriptDirectory.from_config(self.alembic_cfg)
        revs = list(script.revision_map.iterate_revisions(
            'head', revision, select_for_downgrade=True
        ))

        location = self.alembic_cfg.get_main_option('version_locations')

        result = []
        for _, _, filenames in os.walk(location):
            for file in filenames:
                for rev in revs:
                    if rev.revision is not None:
                        # This depends on file_template remaining as it current is in __init__
                        # if that changes, we need a regex for this too
                        if file.startswith(rev.revision):
                            result.append(file)
                            if len(result) == len(revs):
                                return result
                            break
        return result

    def get_history(self):
        script = ScriptDirectory.from_config(self.alembic_cfg)
        return list(script.walk_revisions())

    # Simulates running the `alembic history` command
    def history(self):
        command.history(self.alembic_cfg, indicate_current=True)

    # Simulates running the `alembic current` command
    def current(self):
        command.current(self.alembic_cfg)

    # Simulates running the `alembic show` command
    def show(self, revision):
        command.show(self.alembic_cfg, revision)

    # Simulates running the `alembic heads` command
    def heads(self):
        command.heads(self.alembic_cfg, verbose=True)

    # Simulates running the `alembic branches` command
    def branches(self):
        command.branches(self.alembic_cfg, verbose=True)

    # Simulates running the `alembic stamp` command
    def stamp(self, revision):
        command.stamp(self.alembic_cfg, revision)

    # Simulates running the `alembic edit` command
    # this should probably not be exposed at the moment?
    def edit(self, revision):
        command.edit(self.alembic_cfg, revision)

    def merge(self, revisions, message=None):
        command.merge(self.alembic_cfg, revisions, message=message)

    def squash(self, gen_revision, squash):
        squash = int(squash)
        if squash < 2:
            raise ValueError("squash needs to be an integer of at least 2")

        # downgrade -2 and re-run upgrade
        self.downgrade('-%d' % squash)

        # generate a new revision
        gen_revision()
        self.upgrade()



from alembic.operations import Operations, MigrateOperation
import datetime
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.schema import Table
from sqlalchemy.sql.sqltypes import String

from . import ops


def add_edges_from(connection: sa.engine.Connection, edges):
    t = datetime.datetime.now()
    table = _get_table(connection)

    edges_to_write = []
    for edge in edges:
        edge['created_at'] = t
        edge['updated_at'] = t
        edges_to_write.append(edge)

    dialect = connection.dialect.name
    if dialect == 'sqlite':
        stmt = table.insert().values(
            edges_to_write)
    else:
        stmt = postgresql.insert(table).values(
            edges_to_write).on_conflict_do_nothing()

    connection.execute(stmt)


@ Operations.implementation_for(ops.AddEdgesOp)
def add_edges(operations: ops.Operations, operation: ops.AddEdgesOp):
    connection = operations.get_bind()
    add_edges_from(connection, operation.edges)


@ Operations.implementation_for(ops.RemoveEdgesOp)
def drop_edge(operations: ops.Operations, operation: ops.RemoveEdgesOp):
    edge_types = [edge['edge_type'] for edge in operation.edges]

    connection = operations.get_bind()
    table = _get_table(connection)
    connection.execute(
        table.delete().where(table.c.edge_type.in_(edge_types))
    )


@ Operations.implementation_for(ops.ModifyEdgeOp)
def modify_edge(operations: ops.Operations, operation: ops.ModifyEdgeOp):
    connection = operations.get_bind()
    table = _get_table(connection)
    t = datetime.datetime.now()

    edge = operation.new_edge

    connection.execute(
        table.update().where(table.c.edge_type == operation.edge_type).values(edge)
    )


def _get_table(connection: sa.engine.Connection, name: String = 'assoc_edge_config'):
    # todo there has to be a better way to do this instead of reflecting again
    metadata = sa.MetaData()
    metadata.reflect(connection)

    return metadata.tables[name]


@ Operations.implementation_for(ops.AddRowsOp)
def add_rows(operations: ops.Operations, operation: ops.AddRowsOp):
    connection = operations.get_bind()
    table = _get_table(connection, name=operation.table_name)

    connection.execute(
        table.insert().values(operation.rows)
    )


@ Operations.implementation_for(ops.RemoveRowsOp)
def remove_rows(operations: ops.Operations, operation: ops.RemoveRowsOp):
    connection = operations.get_bind()
    table = _get_table(connection, name=operation.table_name)
    if len(operation.pkeys) == 1:
        key = operation.pkeys[0]
        keys = [row[key] for row in operation.rows]
        connection.execute(
            table.delete().where(table.c[key].in_(keys))
        )
    else:
        # multiple clauses. send a sql statement for each row
        for row in operation.rows:
            clauses = [table.c[key] == row[key] for key in operation.pkeys]
            connection.execute(
                table.delete().where(sa.sql.expression.and_(*clauses))
            )


@Operations.implementation_for(ops.ModifyRowsOp)
def modify_rows(operations: ops.Operations, operation: ops.ModifyRowsOp):
    connection = operations.get_bind()
    table = _get_table(connection, operation.table_name)

    if len(operation.pkeys) == 1:
        key = operation.pkeys[0]
        for row in operation.rows:
            connection.execute(
                table.update().where(table.c[key] == row[key]).values(row)
            )
    else:
        for row in operation.rows:
            clauses = [table.c[key] == row[key] for key in operation.pkeys]
            connection.execute(
                table.update().where(sa.sql.expression.and_(*clauses)).values(row)
            )


@Operations.implementation_for(ops.AlterEnumOp)
def alter_enum(operations: ops.Operations, operation: ops.AlterEnumOp):
    connection = operations.get_bind()
    if operation.before is None:
        connection.execute(
            "ALTER TYPE %s ADD VALUE '%s'" % (
                operation.enum_name, operation.value)
        )
    else:
        connection.execute(
            "ALTER TYPE %s ADD VALUE '%s' BEFORE '%s'" % (
                operation.enum_name, operation.value, operation.before)
        )


@Operations.implementation_for(ops.AddEnumOp)
def add_enum_type(operations: ops.Operations, operation: ops.AddEnumOp):
    postgresql.ENUM(*operation.values, name=operation.enum_name).create(
        operations.get_bind())


@Operations.implementation_for(ops.DropEnumOp)
def drop_enum_type(operations: ops.Operations, operation: ops.DropEnumOp):
    postgresql.ENUM(*operation.values, name=operation.enum_name).drop(
        operations.get_bind())



from alembic import command
import alembic
from alembic.util.exc import CommandError
from auto_schema import runner
from sqlalchemy.sql.sqltypes import String
import pytest
import sqlalchemy as sa

from . import conftest
from . import testingutils
from typing import List
import os


# there doesn't seem to be an api for this
def get_stamped_alembic_versions(r: runner.Runner):
    return [row['version_num'] for row in r.get_connection().execute('select * from alembic_version')]


def stash_new_files(r: runner.Runner, l: List[String], l2: List[String]):
    ret = {}

    for path in l2:
        if path not in l:
            file = os.path.join(r.get_schema_path(), 'versions', path)
            with open(file) as f:
                ret[file] = f.readlines()
            os.remove(file)
    return ret


def write_stashed_files(stash):
    for file in stash:
        with open(file, 'w') as w:
            w.writelines(stash[file])


def _add_column_to_metadata(metadata: sa.MetaData, col_name: String):
    return _add_columns_to_metadata(metadata, [col_name])


def _add_columns_to_metadata(metadata: sa.MetaData, col_names: List[String]):
    if len(testingutils.get_sorted_tables(metadata)) != 1:
        raise ValueError("only support one table at the moment")

    return conftest.metadata_with_given_cols_added_to_table(
        metadata,
        [sa.Column(col_name, sa.Integer, nullable=True)
         for col_name in col_names],
    )


class CommandTest(object):

    @ pytest.mark.usefixtures("metadata_with_table")
    @ pytest.mark.parametrize(
        'migrate_rev, delete_files, new_file_count, expected_current_head',
        [
            # migrate down 1, delete files, 2 files left, current head is the 1st in history
            ('-1', True, 2, 1),
            # migrate down 1, don't delete files, 3 files left, current head is the 1st in history
            ('-1', False, 3, 1),
            # migrate to base, delete files, no files left, no current head
            ('base', True, 0, None),
            # migrate to base, don't delete files, 3 files left, no current head
            ('base', False, 3, None),
            # migrate down with specific rev given, delete files, 2 files left, current head is the 1st in history
            (0, True, 2, 1),
            # migrate down with specific rev given, delete files, 3 files left, current head is the 1st in history
            (0, False, 3, 1),
            # migrate down with specific rev given, delete files, 1 files left, current head is the 2nd in history
            (1, True, 1, 2),
            # migrate down with specific rev given, delete files, 3 files left, current head is the 2nd in history
            (1, False, 3, 2),
        ]
    )
    def test_downgrade(self, new_test_runner, metadata_with_table, migrate_rev, delete_files, new_file_count, expected_current_head):
        r: runner.Runner = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        r2 = testingutils.recreate_with_new_metadata(
            r, new_test_runner, metadata_with_table, conftest.metadata_with_table_with_index)

        message = r2.revision_message()
        assert message == "add index accounts_first_name_idx to accounts"

        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.assert_num_tables(r2, 2)

        r3 = testingutils.recreate_metadata_fixture(
            new_test_runner, conftest.metadata_with_base_table_restored(), r2)

        message = r3.revision_message()
        assert message == "drop index accounts_first_name_idx from accounts"

        r3.run()
        testingutils.assert_num_files(r3, 3)
        testingutils.assert_num_tables(r3, 2)

        history = r3.cmd.get_history()
        assert len(history) == 3
        assert len(testingutils.get_version_files(r3)) == 3

        current_versions = get_stamped_alembic_versions(r3)
        assert len(current_versions) == 1
        assert history[0].revision == current_versions[0]

        # migrate down to down revision
        if isinstance(migrate_rev, int):
            migrate_rev = history[migrate_rev].down_revision

        # downgrade and delete files
        r3.downgrade(migrate_rev, delete_files)
        assert len(testingutils.get_version_files(r3)) == new_file_count

        current_versions = get_stamped_alembic_versions(r3)
        if expected_current_head == None:
            assert len(current_versions) == 0
        else:
            assert len(current_versions) == 1
            assert current_versions[0] == history[expected_current_head].revision

    @ pytest.mark.usefixtures("metadata_with_table")
    @ pytest.mark.parametrize(
        'merge_branches_while_upgrading',
        [
            # don't want to automatically merge branches because we could be upgrading in production
            # and can't make any changes here
            False,
            True
        ]
    )
    def test_upgrade(self, new_test_runner, metadata_with_table, merge_branches_while_upgrading):
        r: runner.Runner = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        assert len(r.cmd.get_heads()) == 1

        files = testingutils.get_version_files(r)
        assert len(files) == 1
        r2 = testingutils.new_runner_from_old(
            r,
            new_test_runner,
            _add_column_to_metadata(metadata_with_table, 'new_col1'),
        )
        r2.revision()
        files2 = testingutils.get_version_files(r2)
        assert len(files2) == 2

        assert len(r2.cmd.get_heads()) == 1

        stashed = stash_new_files(r, files, files2)

        r3 = testingutils.new_runner_from_old(
            r,
            new_test_runner,
            _add_column_to_metadata(metadata_with_table, 'new_col2'),
        )

        r3.revision()

        files3 = testingutils.get_version_files(r3)
        assert len(files3) == 2

        write_stashed_files(stashed)

        files3b = testingutils.get_version_files(r3)
        assert len(files3b) == 3

        assert len(r3.cmd.get_heads()) == 2

        # multiple heads, trying to use alembic upgrade on its own causes leads to an error
        with pytest.raises(CommandError):
            command.upgrade(r3.cmd.alembic_cfg, 'head')

        assert len(r3.compute_changes()) > 0

        # sucessfully upgrade
        info = r3.upgrade('head', merge_branches_while_upgrading)

        files3c = testingutils.get_version_files(r3)

        if merge_branches_while_upgrading:
            assert info.setdefault('unmerged_branches', None) == None
            assert info.setdefault('merged_and_upgraded_head', None) == True
            # new head
            assert len(r3.cmd.get_heads()) == 1

            # new file created
            assert len(files3c) == 4

        else:
            assert info.setdefault('unmerged_branches', None) == True
            assert info.setdefault('merged_and_upgraded_head', None) == None
            assert len(r3.cmd.get_heads()) == 2

            # no new file created
            assert len(files3c) == 3

        # reflect to reload
        r3.metadata.reflect()

        r4 = testingutils.new_runner_from_old(
            r3,
            new_test_runner,
            _add_column_to_metadata(r3.metadata, 'new_col3'),
        )

        r4.revision()
        files4 = testingutils.get_version_files(r4)

        # merge revision was either created earlier during upgrade
        # or now when the change is happening
        assert len(files4) == 5
        r4.run()

    @ pytest.mark.usefixtures("metadata_with_table")
    @pytest.mark.parametrize(
        'squash_val, files_left, squash_raises',
        [
            (2, 2, False),
            (3, 1, False),
            (1, 3, True),
        ]
    )
    def test_squash(self, new_test_runner, metadata_with_table, squash_val, files_left, squash_raises):
        r: runner.Runner = new_test_runner(metadata_with_table)
        testingutils.run_and_validate_with_standard_metadata_tables(
            r, metadata_with_table)

        new_metadata = _add_column_to_metadata(metadata_with_table, 'new_col1')
        r2 = testingutils.new_runner_from_old(
            r,
            new_test_runner,
            new_metadata
        )
        r2.run()
        testingutils.assert_num_files(r2, 2)
        testingutils.validate_metadata_after_change(r2, new_metadata)

        new_metadata = _add_columns_to_metadata(
            metadata_with_table, ['new_col1', 'new_col2'])
        r3 = testingutils.new_runner_from_old(
            r2,
            new_test_runner,
            new_metadata
        )
        r3.run()
        testingutils.assert_num_files(r3, 3)
        testingutils.validate_metadata_after_change(r3, new_metadata)

        r3.metadata.reflect()
        if squash_raises:
            with pytest.raises(ValueError):
                r3.squash(squash_val)
        else:
            r3.squash(squash_val)

        # should be squashed down to X files after change
        testingutils.assert_num_files(r3, files_left)
        r3.metadata.reflect()
        testingutils.validate_metadata_after_change(r3, new_metadata)


class TestPostgresCommand(CommandTest):
    pass


# class TestSQLiteCommand(CommandTest):
#     pass



import json
from collections.abc import Mapping
from .diff import Diff
from .clause_text import get_clause_text
import sqlalchemy as sa
from sqlalchemy.sql.elements import TextClause

from alembic.migration import MigrationContext
from alembic.autogenerate import produce_migrations
from alembic.autogenerate import render_python_code
from alembic.util.exc import CommandError

from sqlalchemy.dialects import postgresql
import alembic.operations.ops as alembicops

from . import command
from . import config
from . import ops
from . import renderers
from . import compare
from . import ops_impl
from . import util


class Runner(object):
    def __init__(self, metadata, connection, schema_path):
        self.metadata = metadata
        self.schema_path = schema_path
        self.connection = connection

        config.metadata = self.metadata
        config.connection = connection

        self.mc = MigrationContext.configure(
            connection=self.connection,
            # note that any change here also needs a comparable change in env.py
            opts={
                "compare_type": Runner.compare_type,
                "include_object": Runner.include_object,
                "compare_server_default": Runner.compare_server_default,
                "transaction_per_migration": True,
                "render_item": Runner.render_item,
            },
        )
        self.cmd = command.Command(self.connection, self.schema_path)

    @classmethod
    def from_command_line(cls, metadata, args):
        engine = sa.create_engine(args.engine)
        connection = engine.connect()
        metadata.bind = connection
        return Runner(metadata, connection, args.schema)

    @classmethod
    def fix_edges(cls, metadata, args):
        if isinstance(args, Mapping) and args.get('connection'):
            connection = args.get('connection')
        else:
            engine = sa.create_engine(args.engine)
            connection = engine.connect()

        edges_map = metadata.info.setdefault("edges", {})
        ops_impl.add_edges_from(connection, list(edges_map['public'].values()))

    @classmethod
    def exclude_tables(cls):
        # prevent default POSTGIS tables from affecting this
        return "geography_columns,geometry_columns,raster_columns,raster_overviews,spatial_ref_sys"

    @classmethod
    def compare_type(cls, context, inspected_column, metadata_column, inspected_type, metadata_type):
        # return False if the metadata_type is the same as the inspected_type
        # or None to allow the default implementation to compare these
        # types. a return value of True means the two types do not
        # match and should result in a type change operation.

        # print(context, inspected_column, metadata_column, inspected_type, metadata_type, type(inspected_type), type(metadata_type))

        # going from VARCHAR to Text is accepted && makes sense and we should accept that change.
        if isinstance(inspected_type, sa.VARCHAR) and isinstance(metadata_type, sa.Text):
            return True

        # going from Date() to TIMESTAMP is accepted && makes sense and we should accept that change.
        # TODO: we should allow going from all less-precise to more-precise since we're not losing any information
        if isinstance(inspected_type, sa.Date) and isinstance(metadata_type, sa.TIMESTAMP):
            return True

        return False

    @classmethod
    def include_object(cls, object, name, type, reflected, compare_to):
        exclude_tables = Runner.exclude_tables().split(',')

        if type == "table" and name in exclude_tables:
            return False
        else:
            return True

    @classmethod
    def convert_postgres_boolean(cls, metadata_default):
        boolean_map = {
            'true': 'true',
            't': 'true',
            'y': 'true',
            'yes': 'true',
            'TRUE': 'true',
            '1': 'true',
            'false': 'false',
            'f': 'false',
            'n': 'false',
            'no': 'false',
            'FALSE': 'false',
            '0': 'false',
        }
        # postgres is very liberal about what it allows for boolean values so support that here
        if metadata_default in boolean_map:
            return boolean_map[metadata_default]
        return metadata_default

    @classmethod
    def compare_server_default(cls, context, inspected_column, metadata_column, inspected_default, metadata_default, rendered_metadata_default):
        # let's just do simple comparison for now.
        # Things may get more complicated in the future but this works for now

        new_inspected_default = get_clause_text(inspected_default)
        new_metadata_default = get_clause_text(metadata_default)
        if isinstance(metadata_column.type, sa.Boolean):
            new_inspected_default = cls.convert_postgres_boolean(
                new_inspected_default)
            new_metadata_default = cls.convert_postgres_boolean(
                new_metadata_default)

        if new_inspected_default != new_metadata_default:
            # specific case. not sure why this is needed
            # can be generalized at some point in the future
            if isinstance(new_inspected_default, str) and new_inspected_default.startswith("nextval") and metadata_default is None:
                return False
            return True
        return False

    @classmethod
    def render_item(cls, type_, item, autogen_context):
        def server_default():
            # For some reason, the default rendering is not doing this correctly, so we need to
            # customize the rendering so that this is handled correctly and it adds the sa.text part
            if isinstance(item, TextClause):
                return "sa.text('%s')" % (item.text)
            return False

        def enum():
            if isinstance(item, postgresql.ENUM):
                # render postgres with create_type=False so that the type is not automatically created
                # we want an explicit create type and drop type
                # which we apparently don't get by default
                return "postgresql.ENUM(%s, name='%s', create_type=False)" % (util.render_list_csv(item.enums), item.name)
            return False

        type_map = {
            'server_default': server_default,
            'type': enum,
        }

        if type_ in type_map:
            return type_map[type_]()

        return False

    def get_schema_path(self):
        return self.schema_path

    def get_metadata(self):
        return self.metadata

    def get_connection(self):
        return config.connection

    def compute_changes(self):
        migrations = produce_migrations(self.mc, config.metadata)
        return migrations.upgrade_ops.ops

    def run(self):
        diff = self.compute_changes()

        if len(diff) == 0:
            return None
        else:
            return self._apply_changes(diff)

    def _apply_changes(self, diff):
        # pprint.pprint(diff, indent=2, width=20)

        # migration_script = produce_migrations(self.mc, self.metadata)
        # print(render_python_code(migration_script.upgrade_ops))

        try:
            self.revision(diff)
        except CommandError as err:
            # expected when not up to date so let's make it up to date by upgrading
            if str(err) == "Target database is not up to date.":
                pass
            else:
                raise err

        return self.upgrade()

    def revision_message(self, diff=None):
        if diff is None:
            migrations = produce_migrations(self.mc, config.metadata)
            diff = migrations.upgrade_ops.ops

        d = Diff(diff, group_by_table=False)
        changes = [c["desc"] for c in d.list_changes()]

        message = "\n".join(changes)
        return message

    def revision(self, diff=None):
        # print(self.cmd.current())

        # if diff is None:
        #   diff = self.compute_changes()

        message = self.revision_message(diff)

        self.cmd.revision(message)

        # understand diff and make changes as needed
        # pprint.pprint(migrations, indent=2, width=30)

    def upgrade(self, revision='head', merge_branches=False):
        return self.cmd.upgrade(revision, merge_branches)

    def downgrade(self, revision, delete_files):
        self.cmd.downgrade(revision, delete_files=delete_files)

    def history(self):
        self.cmd.history()

    def current(self):
        self.cmd.current()

    def show(self, revision):
        self.cmd.show(revision)

    def heads(self):
        self.cmd.heads()

    def branches(self):
        self.cmd.branches()

    def stamp(self, revision):
        self.cmd.stamp(revision)

    def edit(self, revision):
        self.cmd.edit(revision)

    def changes(self):
        diff = self.compute_changes()
        d = Diff(diff, group_by_table=True)
        print(json.dumps(d.changes()))

    def merge(self, revisions, message=None):
        self.cmd.merge(revisions, message=message)

    def squash(self, squash):
        self.cmd.squash(self.revision, squash)



from __future__ import annotations
from sqlalchemy.sql.sqltypes import String
from .clause_text import get_clause_text
from .change_type import ChangeType
from .ops import MigrateOpInterface

import alembic.operations.ops as alembicops

from typing import List, Sequence, TypedDict


class Change(TypedDict):
    change: ChangeType
    desc: String
    col: String


class Diff(object):

    def __init__(self: Diff, diff: Sequence[alembicops.MigrateOperation], group_by_table=True):
        self._diff = diff
        self._changes = {}
        self._changes_list = []
        self._group_by_table = group_by_table

        self._exec_ops(self._diff)

    def list_changes(self: Diff) -> List[Change]:
        return self._changes_list

    def changes(self: Diff):
        return self._changes

    def _exec_ops(self: Diff, ops: Sequence[alembicops.MigrateOperation]):
        class_name_map = {
            'CreateTableOp': self._create_table,
            'DropTableOp': self._drop_table,
            'ModifyTableOps': self._modify_table,
            'AddColumnOp': self._add_column,
            'DropColumnOp': self._drop_column,
            'CreateIndexOp': self._create_index,
            'DropIndexOp': self._drop_index,
            'CreateForeignKeyOp': self._create_foreign_key,
            'AlterColumnOp': self._alter_column,
            'CreateUniqueConstraintOp': self._create_unique_constraint,
            # this is still sadly used
            'DropConstraintOp': self._drop_constraint,
            'CreateCheckConstraintOp': self._create_check_constraint,
        }

        for op in ops:
            if isinstance(op, MigrateOpInterface):
                self._custom_migrate_op(op)
            else:
                class_name_map[type(op).__name__](op)

    def _custom_migrate_op(self: Diff, op: MigrateOpInterface):
        self._append_change(op.get_table_name(), {
            "change": op.get_change_type(),
            "desc": op.get_revision_message(),
        })

    def _create_table(self: Diff, op: alembicops.CreateTableOp):
        self._set_change(op.table_name,                 {
            "change": ChangeType.ADD_TABLE,
            "desc": 'add %s table' % op.table_name,
        })

    def _drop_table(self: Diff, op: alembicops.DropTableOp):
        self._set_change(op.table_name,                 {
            "change": ChangeType.DROP_TABLE,
            "desc": 'drop %s table' % op.table_name,
        })

    def _modify_table(self: Diff, op: alembicops.ModifyTableOps):
        self._exec_ops(op.ops)

    def _add_column(self: Diff, op: alembicops.AddColumnOp):
        self._append_change(op.table_name, {
            "change": ChangeType.ADD_COLUMN,
            "desc": "add column %s to table %s" % (op.column.name, op.table_name),
            "col": op.column.name,
        })

    def _drop_column(self: Diff, op: alembicops.DropColumnOp):
        self._append_change(op.table_name, {
            "change": ChangeType.DROP_COLUMN,
            "desc": "drop column %s from table %s" % (op.column_name, op.table_name),
            "col": op.column_name,
        })

    def _create_index(self: Diff, op: alembicops.CreateIndexOp):
        self._append_change(op.table_name, {
            "change": ChangeType.CREATE_INDEX,
            "desc": 'add index %s to %s' % (op.index_name, op.table_name),
        })

    def _drop_index(self: Diff, op: alembicops.DropIndexOp):
        self._append_change(op.table_name, {
            "change": ChangeType.DROP_INDEX,
            "desc": 'drop index %s from %s' % (op.index_name, op.table_name),
        })

    def _create_foreign_key(self: Diff, op: alembicops.CreateForeignKeyOp):
        self._append_change(op.source_table, {
            "change": ChangeType.CREATE_FOREIGN_KEY,
            "desc": 'create fk constraint %s on %s' % (op.constraint_name, op.source_table),
        })

    def _create_unique_constraint(self: Diff, op: alembicops.CreateUniqueConstraintOp):
        self._append_change(op.table_name, {
            "change": ChangeType.CREATE_UNIQUE_CONSTRAINT,
            "desc": 'add unique constraint %s' % op.constraint_name,
        })

    def _drop_constraint(self: Diff, op: alembicops.DropConstraintOp):
        self._append_change(op.table_name, {
            "change": ChangeType.DROP_CHECK_CONSTRAINT,
            "desc": 'drop constraint %s from %s' % (op.constraint_name, op.table_name)
        })

    def _create_check_constraint(self: Diff, op: alembicops.CreateCheckConstraintOp):
        self._append_change(op.table_name, {
            "change": ChangeType.CREATE_CHECK_CONSTRAINT,
            "desc": 'add constraint %s to %s' % (op.constraint_name, op.table_name)
        })

    def _alter_column(self: Diff, op: alembicops.AlterColumnOp):
        def get_desc(op: alembicops.AlterColumnOp):
            if op.modify_type is not None:
                return 'modify column %s type from %s to %s' % (op.column_name, op.existing_type, op.modify_type)
            elif op.modify_nullable is not None:
                return 'modify nullable value of column %s from %s to %s' % (op.column_name, op.existing_nullable, op.modify_nullable)
            elif op.modify_server_default is not None:
                return 'modify server_default value of column %s from %s to %s' % (
                    op.column_name,
                    get_clause_text(op.existing_server_default),
                    get_clause_text(op.modify_server_default))
            elif op.modify_comment:
                return "modify comment of column %s"
            elif op.modify_name:
                return "modify name of column %s"
            elif op.modify_server_default is None and op.existing_server_default is not None:
                return 'modify server_default value of column %s from %s to None' % (
                    op.column_name,
                    get_clause_text(op.existing_server_default)
                )
            else:
                raise ValueError("unsupported alter_column op")

        self._append_change(op.table_name, {
            "change": ChangeType.ALTER_COLUMN,
            "desc": get_desc(op),
        })

    def _append_change(self: Diff, table_name: String, change: Change):
        if not self._group_by_table:
            self._changes_list.append(change)
            return

        changes = []
        if table_name in self._changes:
            changes = self._changes[table_name]
        changes.append(change)
        self._changes[table_name] = changes

    # throws if there's changes already set for a table
    def _set_change(self: Diff, table_name: String, change: Change):
        if not self._group_by_table:
            self._changes_list.append(change)
            return

        if table_name not in self._changes:
            self._changes[table_name] = [
                change
            ]
        else:
            # TODO: this error needs to be better
            # https://github.com/lolopinto/ent/issues/499
            raise Exception(
                "there shouldn't already be an entry for table %s" % table_name)




from logging.config import dictConfig

from sqlalchemy import engine_from_config
from sqlalchemy import pool

from alembic import context

# env.py apparently cannot be relative because of how it's loaded.
from auto_schema import config
from auto_schema import runner

# these 4 needed for custom ops (edges, enums etc)
from auto_schema import ops
from auto_schema import renderers
from auto_schema import compare
from auto_schema import ops_impl

# set a bunch of loggic parameters based on default info in `alembic init as of 6/15/2019`
log_config = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            # 'args': '(sys.stderr,)',
            'stream': 'ext://sys.stdout',
            'level': 'NOTSET',
            'formatter': 'generic'
        }
    },
    'loggers': {
        'root': {
            'level': 'WARN',
            'handlers': ['console'],
            'qualname': ''
        },
        'sqlalchemy': {
            'level': 'WARN',
            'handlers': '',
            'qualname': 'sqlalchemy.engine'
        },
        'alembic': {
            'level': 'INFO',
            'handlers': '',
            'qualname': 'alembic'
        }
    },
    'formatters': {
        'generic': {
            'format': '%(levelname)-5.5s [%(name)s] %(message)s',
            'datefmt': '%H:%M:%S'
        }
    }
}
dictConfig(log_config)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = config.metadata

# connection engine...
#engine = config.engine
connection = config.connection

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.

# set the target metadata that we'll use here
# def set_metadata(metadata):
#     target_metadata = metadata

# def set_engine(engine):
#     engine = engine


# TODO this hasn't been fully tested or doesn't work...
def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        compare_type=runner.Runner.compare_type,
        include_object=runner.Runner.include_object,
        compare_server_default=runner.Runner.compare_server_default,
        render_item=runner.Runner.render_server_default,
        # transaction_per_migration doesn't seem to apply offline
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        compare_type=runner.Runner.compare_type,
        include_object=runner.Runner.include_object,
        compare_server_default=runner.Runner.compare_server_default,
        render_item=runner.Runner.render_item,
        transaction_per_migration=True
    )

    with context.begin_transaction():
        context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import os
from typing import List
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql
from auto_schema.clause_text import get_clause_text
from auto_schema import runner
from sqlalchemy.sql.sqltypes import String
from . import conftest


def assert_num_files(r: runner.Runner, expected_count):
    assert len(get_version_files(r)) == expected_count


def get_version_files(r: runner.Runner):
    versions_path = os.path.join(r.get_schema_path(), "versions")

    if os.path.isdir(versions_path):
        files = os.listdir(versions_path)
        files = list(filter(lambda f: f != '__pycache__', files))
        return files
    return []


# get sorted tabels not alembic_version
def get_sorted_tables(metadata: sa.MetaData) -> List[sa.Table]:
    return list(
        filter(lambda t: t.name != 'alembic_version', metadata.sorted_tables))


def assert_num_tables(r: runner.Runner, expected_count, tables=None):
    new_metadata = _get_new_metadata_for_runner(r)

    # sometimes, alembic_version is created in between revisions, we don't care about that case here
    # so just write it away
    sorted_tables = new_metadata.sorted_tables
    if expected_count == 0:
        sorted_tables = get_sorted_tables(new_metadata)

    assert len(sorted_tables) == expected_count

    if expected_count > 0 and tables is not None:
        table_names = [t.name for t in sorted_tables]
        table_names.sort()
        assert table_names == tables


def assert_no_changes_made(r: runner.Runner):
    assert_num_files(r, 0)
    assert_num_tables(r, 0)


def validate_edges_from_metadata(metadata: sa.MetaData, r: runner.Runner):
    edges_from_metadata = metadata.info.setdefault("edges", {})

    # denestify the schema
    if len(edges_from_metadata) != 0:
        edges_from_metadata = edges_from_metadata['public']

    db_edges = {}
    for row in r.get_connection().execute("SELECT * FROM assoc_edge_config"):
        row_dict = dict(row)
        db_edges[row_dict['edge_name']] = row_dict

    # same number of edges
    assert len(db_edges) == len(edges_from_metadata)

    for k, edge in edges_from_metadata.items():
        db_edge = db_edges.get(k)

        assert db_edge is not None

        assert db_edge['edge_name'] == edge['edge_name']
        assert db_edge['edge_type'] == edge['edge_type']
        assert db_edge['edge_table'] == edge['edge_table']
        assert db_edge.get('inverse_edge_type') == edge.get(
            'inverse_edge_type')
        # 0 == False??
        assert db_edge.get('symmetric_edge') == edge.get('symmetric_edge')


def validate_data_from_metadata(metadata: sa.MetaData, r: runner.Runner):
    def sort_rows(rows):
        # sort each time by pkey and depend on the fact that sorting is stable
        for pkey in pkeys:
            rows.sort(key=lambda obj: obj[pkey])

    data_from_metadata = metadata.info.setdefault('data', {})
    if len(data_from_metadata) != 0:
        data_from_metadata = data_from_metadata['public']

    for table_name in data_from_metadata:

        data_rows = data_from_metadata[table_name]['rows']
        pkeys = data_from_metadata[table_name]['pkeys']
        sort_rows(data_rows)

        db_rows = []
        db_keys = []
        for row in r.get_connection().execute('SELECT * FROM %s' % table_name):
            row_dict = dict(row)
            if len(db_keys) == 0:
                db_keys = row_dict.keys()
            db_rows.append(row_dict)

        sort_rows(db_rows)

        # verify data in db is same after sorting
        # we go through each data row and set None for any missing keys
        for index, row in enumerate(data_rows):
            # set None for nullable keys that aren't specified in data rows
            for key in db_keys:
                row[key] = row.get(key, None)
            data_rows[index] = row

        assert data_rows == db_rows


def validate_metadata_after_change(r: runner.Runner, old_metadata: sa.MetaData):
    new_metadata = _get_new_metadata_for_runner(r)
    assert new_metadata != old_metadata

    dialect = r.get_connection().dialect.name
# TODO why is this here?
#    assert(len(old_metadata.sorted_tables)) != len(new_metadata.sorted_tables)

    new_metadata.bind = r.get_connection()
    for db_table in new_metadata.sorted_tables:
        schema_table = next(
            (t for t in old_metadata.sorted_tables if db_table.name == t.name), None)

        if schema_table is not None:
            _validate_table(schema_table, db_table, dialect, new_metadata)
        else:
            # no need to do too much testing on this since we'll just have to trust that alembic works.
            assert db_table.name == 'alembic_version'


def run_and_validate_with_standard_metadata_tables(r: runner.Runner, metadata_with_table: sa.MetaData, new_table_names=['accounts']):
    r.run()

    # should have the expected file with the expected tables
    assert_num_files(r, 1)
    tables = ['alembic_version']
    [tables.append(t) for t in new_table_names]
    tables.sort()
    assert_num_tables(r, len(tables), tables)

    validate_metadata_after_change(r, metadata_with_table)


def recreate_with_new_metadata(r: runner.Runner, new_test_runner, metadata_with_table: sa.MetaData, metadata_func) -> runner.Runner:
    metadata_func(metadata_with_table)
    # recreate and run
    r2 = new_test_runner(metadata_with_table, r)
    assert r.get_schema_path() == r2.get_schema_path()
    return r2


# TODO too many of these functions and not obvious what the difference is
def new_runner_from_old(prev_runner: runner.Runner, new_test_runner, new_metadata):
    new_metadata.bind = prev_runner.get_connection()
    r2 = new_test_runner(new_metadata, prev_runner)
    return r2


def recreate_metadata_fixture(new_test_runner, metadata: sa.MetaData, prev_runner: runner.Runner) -> runner.Runner:
    metadata.bind = prev_runner.get_connection()
    metadata.reflect()

    r = new_test_runner(metadata, prev_runner)
    return r


def run_edge_metadata_script(new_test_runner, metadata: sa.MetaData, message: String, num_files=2, prev_runner=None, num_changes=1) -> runner.Runner:
    # TODO combine with recreate_with_new_metadata?
    if prev_runner is None:
        prev_runner = _setup_assoc_edge_config(new_test_runner)

    r = recreate_metadata_fixture(new_test_runner, metadata, prev_runner)
    assert len(r.compute_changes()) == num_changes

    assert r.revision_message() == message

    r.run()
    # new file added for edge
    assert_num_files(r, num_files)
    validate_edges_from_metadata(metadata, r)

    return r


def _get_new_metadata_for_runner(r: runner.Runner) -> sa.MetaData:
    # metadata = r.get_metadata()
    # don't reflect but in fact get a new object so that we can reflect corectly
    new_metadata = sa.MetaData()
    # fetch any new tables
    new_metadata.reflect(bind=r.get_connection())
    return new_metadata


def _validate_table(schema_table: sa.Table, db_table: sa.Table, dialect: String, metadata: sa.MetaData):
    assert schema_table != db_table
    assert id(schema_table) != id(db_table)

    assert schema_table.name == db_table.name

    _validate_columns(schema_table, db_table, metadata, dialect)
    _validate_constraints(schema_table, db_table, dialect, metadata)
    _validate_indexes(schema_table, db_table, metadata)


def _validate_columns(schema_table: sa.Table, db_table: sa.Table, metadata: sa.MetaData, dialect: String):
    schema_columns = schema_table.columns
    db_columns = db_table.columns
    assert len(schema_columns) == len(db_columns)
    for schema_column, db_column in zip(schema_columns, db_columns):
        _validate_column(schema_column, db_column, metadata, dialect)


def _validate_column(schema_column: sa.Column, db_column: sa.Column, metadata: sa.MetaData, dialect: String):
    assert schema_column != db_column
    assert(id(schema_column)) != id(db_column)

    assert schema_column.name == db_column.name
    _validate_column_type(schema_column, db_column, metadata, dialect)
    assert schema_column.primary_key == db_column.primary_key
    assert schema_column.nullable == db_column.nullable

    _validate_foreign_key(schema_column, db_column)
    _validate_column_server_default(schema_column, db_column)

    # we don't actually support all these below yet but when we do, it should start failing and we should know that
    assert schema_column.default == db_column.default
    assert schema_column.index == db_column.index
    assert schema_column.unique == db_column.unique
    # assert schema_column.autoincrement == db_column.autoincrement # ignore autoincrement for now as there's differences btw default behavior and postgres
    assert schema_column.key == db_column.key
    assert schema_column.onupdate == db_column.onupdate
    assert schema_column.constraints == db_column.constraints
    assert len(schema_column.constraints) == 0
    assert schema_column.comment == db_column.comment


def _validate_column_server_default(schema_column: sa.Column, db_column: sa.Column):
    schema_clause_text = get_clause_text(
        schema_column.server_default)
    db_clause_text = get_clause_text(db_column.server_default)

    if isinstance(schema_column.type, sa.Boolean):
        schema_clause_text = runner.Runner.convert_postgres_boolean(
            schema_clause_text)
        db_clause_text = runner.Runner.convert_postgres_boolean(db_clause_text)

    if schema_clause_text is None and db_column.autoincrement == True:
        assert db_clause_text.startswith("nextval")
    else:
        assert schema_clause_text == db_clause_text


def _validate_column_type(schema_column: sa.Column, db_column: sa.Column, metadata: sa.MetaData, dialect: String):
    # array type. validate contents
    if isinstance(schema_column.type, postgresql.ARRAY):
        assert isinstance(db_column.type, postgresql.ARRAY)

        _validate_column_type_impl(
            schema_column.type.item_type, db_column.type.item_type, metadata, dialect, db_column, schema_column)
    else:

        _validate_column_type_impl(
            schema_column.type, db_column.type, metadata, dialect, db_column, schema_column)
    pass


def _validate_column_type_impl(schema_column_type, db_column_type, metadata: sa.MetaData, dialect, db_column: sa.Column, schema_column: sa.Column):

    if isinstance(schema_column_type, sa.TIMESTAMP):
        # timezone not supported in sqlite so this is just ignored there
        if dialect != 'sqlite':
            assert schema_column_type.timezone == db_column_type.timezone
        else:
            assert str(db_column_type) == "TIMESTAMP"
    elif isinstance(schema_column_type, sa.Time):
        # timezone not supported in sqlite so this is just ignored there
        if dialect != 'sqlite':
            assert schema_column_type.timezone == db_column_type.timezone
        else:
            assert str(db_column_type) == "TIME"
    elif isinstance(schema_column_type, sa.Numeric):
        assert isinstance(db_column_type, sa.Numeric)
        # precision is tricky so ignore this for now
        # assert schema_column.type.precision == db_column.type.precision
    elif isinstance(schema_column_type, postgresql.ENUM):
        # enum type if possible otherwise check constraint...
        assert isinstance(db_column_type, postgresql.ENUM)
        _validate_enum_column_type(metadata, db_column, schema_column)
    else:
        # compare types by using the string version of the types.
        # seems to account for differences btw Integer and INTEGER, String(255) and VARCHAR(255) etc

        assert str(schema_column_type) == str(db_column_type)


def _validate_enum_column_type(metadata: sa.MetaData, db_column: sa.Column, schema_column: sa.Column):
    # has to be same length
    assert(len(schema_column.type.enums) == len(db_column.type.enums))

    # if equal, nothing to do here, we're done
    if schema_column.type.enums == db_column.type.enums:
        return

    # we gotta go to the db and check the order
    db_sorted_enums = []
    # https://www.postgresql.org/docs/9.5/functions-enum.html
    query = "select unnest(enum_range(enum_first(null::%s)));" % (
        db_column.type.name)
    for row in metadata.bind.execute(query):
        db_sorted_enums.append(dict(row)['unnest'])

    assert schema_column.type.enums == db_sorted_enums


def _sort_fn(item):
    # if name is null, use type of object to sort
    if item.name is None:
        return type(item).__name__ + str(id(item))
    # otherwise, use name + class name
    return type(item).__name__ + item.name


def _validate_indexes(schema_table: sa.Table, db_table: sa.Table, metadata: sa.MetaData):
    # sort indexes so that the order for both are the same
    schema_indexes = sorted(schema_table.indexes, key=_sort_fn)
    db_indexes = sorted(db_table.indexes, key=_sort_fn)

    assert len(schema_indexes) == len(db_indexes)
    for schema_index, db_index in zip(schema_indexes, db_indexes):
        # index names should be equal
        assert schema_index.name == db_index.name

        schema_index_columns = schema_index.columns
        db_index_columns = db_index.columns
        for schema_column, db_column in zip(schema_index_columns, db_index_columns):
            _validate_column(schema_column, db_column, metadata)


def _validate_constraints(schema_table: sa.Table, db_table: sa.Table, dialect: String, metadata: sa.MetaData):
    # sort constraints so that the order for both are the same
    schema_constraints = sorted(schema_table.constraints, key=_sort_fn)
    db_constraints = sorted(db_table.constraints, key=_sort_fn)

    bool_column_names_set = set()
    # sqlite doesn't support native boolean datatype so it adds another constraint.
    # This is us working around that...
    if dialect == 'sqlite':
        # remove the extra sqlite specific boolean constraints
        bool_columns = list(filter(lambda col: str(
            col.type) == 'BOOLEAN', db_table.columns))
        bool_columns_set = set(bool_columns)
        bool_column_names_set = set([col.name for col in bool_columns])
        db_constraints_to_compare = []
        for constraint in db_constraints:
            if isinstance(constraint, sa.CheckConstraint) and len(constraint.columns) == 1 and len(bool_columns_set.intersection(constraint.columns)) > 0:
                continue
            db_constraints_to_compare.append(constraint)

        db_constraints = db_constraints_to_compare

    assert len(schema_constraints) == len(db_constraints)

    for schema_constraint, db_constraint in zip(schema_constraints, db_constraints):
        # constraint names should be equal
        if schema_constraint.name == '_unnamed_' and dialect == 'sqlite':
            assert db_constraint.name == None
        else:
            assert schema_constraint.name == db_constraint.name

        schema_constraint_columns = schema_constraint.columns
        db_constraint_columns = db_constraint.columns

        if (dialect == 'sqlite' and
            isinstance(db_constraint, sa.CheckConstraint) and
            len(schema_constraint_columns) == 1 and
            len(db_constraint_columns) == 0 and
            # sqlalchemy's default check constraint adds an extra rule for dialects that don't natively
            # support booleans, if we're in this case, check for it and don't do the rest of the checks
            # see _should_create_constraint in sqltypes.py
                len(bool_column_names_set.intersection([col.name for col in schema_constraint_columns])) == 1):
            continue

        assert len(schema_constraint_columns) == len(db_constraint_columns)
        for schema_column, db_column in zip(schema_constraint_columns, db_constraint_columns):
            _validate_column(schema_column, db_column, metadata, dialect)


def _validate_foreign_key(schema_column: sa.Column, db_column: sa.Column):
    assert len(schema_column.foreign_keys) == len(schema_column.foreign_keys)

    for db_fkey, schema_fkey in zip(db_column.foreign_keys, schema_column.foreign_keys):
        # similar to what we do in validate_table on column.type
        assert str(db_fkey.column) == str(schema_fkey.column)
        assert db_fkey.name == schema_fkey.name
        assert db_fkey.ondelete == schema_fkey.ondelete
        assert db_fkey.onupdate == schema_fkey.onupdate

        # we don't actually support all these below yet but when we do, it should start failing and we should know that
        assert db_fkey.deferrable == schema_fkey.deferrable
        assert db_fkey.initially == schema_fkey.initially
       # assert db_fkey.link_to_name == schema_fkey.link_to_name # this seems like it's expected to change. TODO figure this out more
        assert db_fkey.use_alter == schema_fkey.use_alter
        assert db_fkey.match == schema_fkey.match
        assert db_fkey.info == schema_fkey.info
        assert str(db_fkey.parent) == str(schema_fkey.parent)


def _setup_assoc_edge_config(new_test_runner):
    # no revision, just do the changes to setup base case
    r = new_test_runner(conftest.metadata_assoc_edge_config())
    assert len(r.compute_changes()) == 1
    r.run()
    assert_num_tables(r, 2, ['alembic_version', 'assoc_edge_config'])
    assert_num_files(r, 1)
    # r.get_metadata().reflect(bind=r.get_connection())

    return r



from sqlalchemy.sql.schema import DefaultClause
from sqlalchemy.sql.elements import TextClause


def get_clause_text(server_default):
    if server_default is None:
        return server_default

    def normalize(arg):
        # return the underlying string instead of quoted
        arg = str(arg).strip("'")

        # strip the extra text padding added so we can compare effectively
        if arg.endswith("'::text"):
            return arg.strip("'::text")
        return arg

    if isinstance(server_default, TextClause):
        return normalize(server_default.text)

    if isinstance(server_default, DefaultClause):
        return normalize(server_default.arg)

    return normalize(server_default)



from alembic.autogenerate import renderers
from alembic.autogenerate.api import AutogenContext
from . import ops
from . import util
import sqlalchemy as sa

# no need to put timestamps when rendering
_IGNORED_KEYS = ['created_at', 'updated_at']


@renderers.dispatch_for(ops.AddEdgesOp)
def render_add_edges(autogen_context: AutogenContext, op: ops.AddEdgesOp):
    return _render_edge_from_edges(op.edges, "op.add_edges")


@renderers.dispatch_for(ops.RemoveEdgesOp)
def render_remove_edges(autogen_context: AutogenContext, op: ops.RemoveEdgesOp):
    return _render_edge_from_edges(op.edges, "op.remove_edges")


@renderers.dispatch_for(ops.ModifyEdgeOp)
def render_modify_edge(autogen_context: AutogenContext, op: ops.ModifyEdgeOp):
    return (
        "op.modify_edge(\n"
        "'%(edge_type)s',\n"
        "%(edge)s\n"
        ")" % {
            "edge_type": op.edge_type,
            "edge": _render_edge(op.new_edge),
        }
    )


def _render_edge(edge):
    kv_pairs = []
    # get each line for each edge
    for k, v in edge.items():
        if k in _IGNORED_KEYS:
            continue

        kv_pairs.append("'%s': %r" % (k, v))

    # get the rendering for an edge
    #     {"k": v, "k2": v2}
    return "{%s},\n" % ", ".join(kv_pairs)


def _render_edge_from_edges(edge_dicts, edge_fn_name):
    # build each edge
    edges = []
    for edge in edge_dicts:
        # append the code for each edge into list. we should end with something like:
        edges.append(
            _render_edge(edge)
        )

    # splice the edges to be rendered
    return (
        "%(edge_fn_name)s([\n"
        "%(edges)s"
        "])\n" % {
            "edge_fn_name": edge_fn_name,
            "edges": "".join(edges),
        }
    )


def _render_row_from_op(row_fn_name, table_name, pkeys, rows):
    rows = [_render_row(row) for row in rows]

    # splice the rows to be rendered
    return (
        "%(row_fn_name)s('%(table_name)s', %(pkeys)s, [\n"
        "%(rows)s"
        "])" % {
            "row_fn_name": row_fn_name,
            "table_name": table_name,
            "pkeys": util.render_list_csv_as_list(pkeys),
            "rows": "".join(rows),
        }
    )


def _render_row(row):
    kv_pairs = []
    # get each line for each row
    for k, v in row.items():
        kv_pairs.append("'%s': %r" % (k, v))

    # get the rendering for a row
    #     {"k": v, "k2": v2}
    return "{%s},\n" % ", ".join(kv_pairs)


@renderers.dispatch_for(ops.AddRowsOp)
def render_add_edges(autogen_context: AutogenContext, op: ops.AddRowsOp):
    return _render_row_from_op("op.add_rows", op.table_name, op.pkeys, op.rows)


@renderers.dispatch_for(ops.RemoveRowsOp)
def render_remove_edges(autogen_context: AutogenContext, op: ops.RemoveRowsOp):
    return _render_row_from_op("op.remove_rows", op.table_name, op.pkeys, op.rows)


@renderers.dispatch_for(ops.ModifyRowsOp)
def render_modify_rows(autogen_context: AutogenContext, op: ops.ModifyRowsOp):
    rows = [_render_row(row) for row in op.rows]
    old_rows = [_render_row(row) for row in op.old_rows]

    return (
        "op.modify_rows('%(table_name)s', %(pkeys)s, [\n"
        "%(rows)s],"
        "[\n%(old_rows)s"
        "])" % {
            "table_name": op.table_name,
            "pkeys": util.render_list_csv_as_list(op.pkeys),
            "rows": "".join(rows),
            "old_rows": "".join(old_rows),
        }
    )


@renderers.dispatch_for(ops.AlterEnumOp)
def render_alter_enum(autogen_context: AutogenContext, op: ops.AlterEnumOp):
    if op.before is None:
        return (
            # manual indentation
            "with op.get_context().autocommit_block():\n"
            "    op.alter_enum('%(enum_name)s', '%(value)s')" % {
                "enum_name": op.enum_name,
                "value": op.value,
            }
        )
    else:
        return (
            # manual indentation
            "with op.get_context().autocommit_block():\n"
            "    op.alter_enum('%(enum_name)s', '%(value)s', before='%(before)s')" % {
                "enum_name": op.enum_name,
                "value": op.value,
                "before": op.before,
            }
        )


@renderers.dispatch_for(ops.NoDowngradeOp)
def render_no_downgrade_op(autogen_context: AutogenContext, op: ops.NoDowngradeOp):
    return "raise ValueError('operation is not reversible')"


@renderers.dispatch_for(ops.AddEnumOp)
def render_add_enum_op(autogen_context: AutogenContext, op: ops.AddEnumOp):
    return "op.add_enum_type('%s', [%s])" % (op.enum_name, util.render_list_csv(op.values))


@renderers.dispatch_for(ops.DropEnumOp)
def render_drop_enum_op(autogen_context: AutogenContext, op: ops.DropEnumOp):
    return "op.drop_enum_type('%s', [%s])" % (op.enum_name, util.render_list_csv(op.values))


# hmm not sure why we need this. there's probably an easier way to do this that doesn't require this...
@renderers.dispatch_for(ops.OurCreateCheckConstraintOp)
def render_check_constraint(autogen_context: AutogenContext, op: ops.OurCreateCheckConstraintOp):
    return "op.create_check_constraint('%s', '%s', '%s')" % (op.constraint_name, op.table_name, op.condition)
